import collections
import io
import itertools
import re
import sys
from argparse import ArgumentParser

import ruamel.yaml

import borgmatic.config.schema
from borgmatic.config import collect

ACTION_ALIASES = {
    'repo-create': ['rcreate', 'init', '-I'],
    'prune': ['-p'],
    'compact': [],
    'create': ['-C'],
    'check': ['-k'],
    'config': [],
    'delete': [],
    'extract': ['-x'],
    'export-tar': [],
    'mount': ['-m'],
    'umount': ['-u'],
    'restore': ['-r'],
    'repo-delete': ['rdelete'],
    'repo-list': ['rlist'],
    'list': ['-l'],
    'repo-info': ['rinfo'],
    'info': ['-i'],
    'transfer': [],
    'break-lock': [],
    'key': [],
    'borg': [],
    'recreate': [],
    'diff': [],
}


def get_subaction_parsers(action_parser):
    '''
    Given an argparse.ArgumentParser instance, lookup the subactions in it and return a dict from
    subaction name to subaction parser.
    '''
    if not action_parser._subparsers:
        return {}

    return {
        subaction_name: subaction_parser
        for group_action in action_parser._subparsers._group_actions
        for subaction_name, subaction_parser in group_action.choices.items()
    }


def get_subactions_for_actions(action_parsers):
    '''
    Given a dict from action name to an argparse.ArgumentParser instance, make a map from action
    name to the names of contained sub-actions.
    '''
    return {
        action: tuple(
            subaction_name
            for group_action in action_parser._subparsers._group_actions
            for subaction_name in group_action.choices
        )
        for action, action_parser in action_parsers.items()
        if action_parser._subparsers
    }


def omit_values_colliding_with_action_names(unparsed_arguments, parsed_arguments):
    '''
    Given unparsed arguments as a sequence of strings and a dict from action name to parsed
    argparse.Namespace arguments, return the string arguments with any values omitted that happen to
    be the same as the name of a borgmatic action.

    This prevents, for instance, "check --only extract" from triggering the "extract" action.
    '''
    remaining_arguments = list(unparsed_arguments)

    for parsed in parsed_arguments.values():
        for value in vars(parsed).values():
            if isinstance(value, str):
                if value in ACTION_ALIASES and value in remaining_arguments:
                    remaining_arguments.remove(value)
            elif isinstance(value, list):
                for item in value:
                    if item in ACTION_ALIASES and item in remaining_arguments:
                        remaining_arguments.remove(item)

    return tuple(remaining_arguments)


def parse_and_record_action_arguments(
    unparsed_arguments,
    parsed_arguments,
    action_parser,
    action_name,
    canonical_name=None,
):
    '''
    Given unparsed arguments as a sequence of strings, parsed arguments as a dict from action name
    to parsed argparse.Namespace, a parser to parse with, an action name, and an optional canonical
    action name (in case this the action name is an alias), parse the arguments and return a list of
    any remaining string arguments that were not parsed. Also record the parsed argparse.Namespace
    by setting it into the given parsed arguments. Return None if no parsing occurs because the
    given action doesn't apply to the given unparsed arguments.
    '''
    filtered_arguments = omit_values_colliding_with_action_names(
        unparsed_arguments,
        parsed_arguments,
    )

    if action_name not in filtered_arguments:
        return tuple(unparsed_arguments)

    parsed, remaining = action_parser.parse_known_args(filtered_arguments)
    parsed_arguments[canonical_name or action_name] = parsed

    # Special case: If this is a "borg" action, greedily consume all arguments after (+1) the "borg"
    # argument.
    if action_name == 'borg':
        borg_options_index = remaining.index('borg') + 1
        parsed_arguments['borg'].options = remaining[borg_options_index:]
        remaining = remaining[:borg_options_index]

    return tuple(argument for argument in remaining if argument != action_name)


def argument_is_flag(argument):
    '''
    Return True if the given argument looks like a flag, e.g. '--some-flag', as opposed to a
    non-flag value.
    '''
    return isinstance(argument, str) and argument.startswith('--')


def group_arguments_with_values(arguments):
    '''
    Given a sequence of arguments, return a sequence of tuples where each one contains either a
    single argument (such as for a stand-alone flag) or a flag argument and its corresponding value.

    For instance, given the following arguments sequence as input:

      ('--foo', '--bar', '33', '--baz')

    ... return the following output:

      (('--foo',), ('--bar', '33'), ('--baz',))
    '''
    grouped_arguments = []
    index = 0

    while index < len(arguments):
        this_argument = arguments[index]

        try:
            next_argument = arguments[index + 1]
        except IndexError:
            grouped_arguments.append((this_argument,))
            break

        if (
            argument_is_flag(this_argument)
            and not argument_is_flag(next_argument)
            and next_argument not in ACTION_ALIASES
        ):
            grouped_arguments.append((this_argument, next_argument))
            index += 2
            continue

        grouped_arguments.append((this_argument,))
        index += 1

    return tuple(grouped_arguments)


def get_unparsable_arguments(remaining_action_arguments):
    '''
    Given a sequence of argument tuples (one per action parser that parsed arguments), determine the
    remaining arguments that no action parsers have consumed.
    '''
    if not remaining_action_arguments:
        return ()

    grouped_action_arguments = tuple(
        group_arguments_with_values(action_arguments)
        for action_arguments in remaining_action_arguments
    )

    return tuple(
        itertools.chain.from_iterable(
            argument_group
            for argument_group in dict.fromkeys(
                itertools.chain.from_iterable(grouped_action_arguments),
            )
            if all(
                argument_group in action_arguments for action_arguments in grouped_action_arguments
            )
        ),
    )


def parse_arguments_for_actions(unparsed_arguments, action_parsers, global_parser):
    '''
    Given a sequence of arguments, a dict from action name to argparse.ArgumentParser instance,
    and the global parser as a argparse.ArgumentParser instance, give each requested action's
    parser a shot at parsing all arguments. This allows common arguments like "--repository" to be
    shared across multiple action parsers.

    Return the result as a tuple of: (a dict mapping from action name to an argparse.Namespace of
    parsed arguments, a tuple of argument tuples where each is the remaining arguments not claimed
    by any action parser).
    '''
    arguments = collections.OrderedDict()
    help_requested = bool('--help' in unparsed_arguments or '-h' in unparsed_arguments)
    remaining_action_arguments = []
    alias_to_action_name = {
        alias: action_name for action_name, aliases in ACTION_ALIASES.items() for alias in aliases
    }

    # If the "borg" action is used, skip all other action parsers. This avoids confusion like
    # "borg list" triggering borgmatic's own list action.
    if 'borg' in unparsed_arguments:
        action_parsers = {'borg': action_parsers['borg']}

    # Ask each action parser, one by one, to parse arguments.
    for argument in unparsed_arguments:
        action_name = argument
        canonical_name = alias_to_action_name.get(action_name, action_name)
        action_parser = action_parsers.get(action_name)

        if not action_parser:
            continue

        subaction_parsers = get_subaction_parsers(action_parser)

        # But first parse with subaction parsers, if any.
        if subaction_parsers:
            subactions_parsed = False

            for subaction_name, subaction_parser in subaction_parsers.items():
                remaining_action_arguments.append(
                    tuple(
                        argument
                        for argument in parse_and_record_action_arguments(
                            unparsed_arguments,
                            arguments,
                            subaction_parser,
                            subaction_name,
                        )
                        if argument != action_name
                    ),
                )

                if subaction_name in arguments:
                    subactions_parsed = True

            if not subactions_parsed:
                if help_requested:
                    action_parser.print_help()
                    sys.exit(0)
                else:
                    raise ValueError(
                        f"Missing sub-action after {action_name} action. Expected one of: {', '.join(get_subactions_for_actions(action_parsers)[action_name])}",
                    )
        # Otherwise, parse with the main action parser.
        else:
            remaining_action_arguments.append(
                parse_and_record_action_arguments(
                    unparsed_arguments,
                    arguments,
                    action_parser,
                    action_name,
                    canonical_name,
                ),
            )

    # If no actions were explicitly requested, assume defaults.
    if not arguments and not help_requested:
        for default_action_name in ('create', 'prune', 'compact', 'check'):
            default_action_parser = action_parsers[default_action_name]
            remaining_action_arguments.append(
                parse_and_record_action_arguments(
                    (*unparsed_arguments, default_action_name),
                    arguments,
                    default_action_parser,
                    default_action_name,
                ),
            )

    arguments['global'], remaining = global_parser.parse_known_args(unparsed_arguments)
    remaining_action_arguments.append(remaining)

    return (
        arguments,
        tuple(remaining_action_arguments) if arguments else unparsed_arguments,
    )


OMITTED_FLAG_NAMES = {
    'match-archives',
    'progress',
    'statistics',
    'list-details',
    'file-list-format',
    'archive-list-format',
}


def make_argument_description(schema, flag_name):
    '''
    Given a configuration schema dict and a flag name for it, extend the schema's description with
    an example or additional information as appropriate based on its type. Return the updated
    description for use in a command-line argument.
    '''
    description = schema.get('description')
    schema_type = schema.get('type')
    example = schema.get('example')
    pieces = [description] if description else []

    if '[0]' in flag_name:
        pieces.append(
            ' To specify a different list element, replace the "[0]" with another array index ("[1]", "[2]", etc.).',
        )

    if example and schema_type in ('array', 'object'):  # noqa: PLR6201
        example_buffer = io.StringIO()
        yaml = ruamel.yaml.YAML(typ='safe')
        yaml.default_flow_style = True
        yaml.dump(example, example_buffer)

        pieces.append(f'Example value: "{example_buffer.getvalue().strip()}"')

    return ' '.join(pieces).replace('%', '%%')


def add_array_element_arguments(arguments_group, unparsed_arguments, flag_name):
    r'''
    Given an argparse._ArgumentGroup instance, a sequence of unparsed argument strings, and a dotted
    flag name, add command-line array element flags that correspond to the given unparsed arguments.

    Here's the background. We want to support flags that can have arbitrary indices like:

      --foo.bar[1].baz

    But argparse doesn't support that natively because the index can be an arbitrary number. We
    won't let that stop us though, will we?

    If the current flag name has an array component in it (e.g. a name with "[0]"), then make a
    pattern that would match the flag name regardless of the number that's in it. The idea is that
    we want to look for unparsed arguments that appear like the flag name, but instead of "[0]" they
    have, say, "[1]" or "[123]".

    Next, we check each unparsed argument against that pattern. If one of them matches, add an
    argument flag for it to the argument parser group. Example:

    Let's say flag_name is:

        --foo.bar[0].baz

    ... then the regular expression pattern will be:

        ^--foo\.bar\[\d+\]\.baz

    ... and, if that matches an unparsed argument of:

        --foo.bar[1].baz

    ... then an argument flag will get added equal to that unparsed argument. And so the unparsed
    argument will match it when parsing is performed! In this manner, we're using the actual user
    CLI input to inform what exact flags we support.
    '''
    if '[0]' not in flag_name or not unparsed_arguments or '--help' in unparsed_arguments:
        return

    pattern = re.compile('^--' + flag_name.replace('[0]', r'\[\d+\]').replace('.', r'\.') + '$')

    try:
        # Find an existing list index flag (and its action) corresponding to the given flag name.
        (argument_action, existing_flag_name) = next(
            (action, action_flag_name)
            for action in arguments_group._group_actions
            for action_flag_name in action.option_strings
            if pattern.match(action_flag_name)
            if f'--{flag_name}'.startswith(action_flag_name)
        )

        # Based on the type of the action (e.g. argparse._StoreTrueAction), look up the corresponding
        # action registry name (e.g., "store_true") to pass to add_argument(action=...) below.
        action_registry_name = next(
            registry_name
            for registry_name, action_type in arguments_group._registries['action'].items()
            # Not using isinstance() here because we only want an exact match—no parent classes.
            if type(argument_action) is action_type
        )
    except StopIteration:
        return

    for unparsed in unparsed_arguments:
        unparsed_flag_name = unparsed.split('=', 1)[0]
        destination_name = unparsed_flag_name.lstrip('-').replace('-', '_')

        if not pattern.match(unparsed_flag_name) or unparsed_flag_name == existing_flag_name:
            continue

        if action_registry_name in {'store_true', 'store_false'}:
            arguments_group.add_argument(
                unparsed_flag_name,
                action=action_registry_name,
                default=argument_action.default,
                dest=destination_name,
                required=argument_action.nargs,
            )
        else:
            arguments_group.add_argument(
                unparsed_flag_name,
                action=action_registry_name,
                choices=argument_action.choices,
                default=argument_action.default,
                dest=destination_name,
                nargs=argument_action.nargs,
                required=argument_action.nargs,
                type=argument_action.type,
            )


def add_arguments_from_schema(arguments_group, schema, unparsed_arguments, names=None):  # noqa: PLR0912
    '''
    Given an argparse._ArgumentGroup instance, a configuration schema dict, and a sequence of
    unparsed argument strings, convert the entire schema into corresponding command-line flags and
    add them to the arguments group.

    For instance, given a schema of:

        {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'object',
                    'properties': {
                        'bar': {'type': 'integer'}
                    }
                }
            }
        }

    ... the following flag will be added to the arguments group:

        --foo.bar

    If "foo" is instead an array of objects, both of the following will get added:

        --foo
        --foo[0].bar

    And if names are also passed in, they are considered to be the name components of an option
    (e.g. "foo" and "bar") and are used to construct a resulting flag.

    Bail if the schema is not a dict.
    '''
    if names is None:
        names = ()

    if not isinstance(schema, dict):
        return

    schema_type = schema.get('type')

    # If this option has multiple types, just use the first one (that isn't "null").
    if isinstance(schema_type, list):
        try:
            schema_type = next(single_type for single_type in schema_type if single_type != 'null')
        except StopIteration:
            raise ValueError(f'Unknown type in configuration schema: {schema_type}')

    # If this is an "object" type, recurse for each child option ("property").
    if schema_type == 'object':
        properties = schema.get('properties')

        # If there are child properties, recurse for each one. But if there are no child properties,
        # fall through so that a flag gets added below for the (empty) object.
        if properties:
            for name, child in properties.items():
                add_arguments_from_schema(
                    arguments_group,
                    child,
                    unparsed_arguments,
                    (*names, name),
                )

            return

    # If this is an "array" type, recurse for each items type child option. Don't return yet so that
    # a flag also gets added below for the array itself.
    if schema_type == 'array':
        items = schema.get('items', {})
        properties = borgmatic.config.schema.get_properties(items)

        if properties:
            for name, child in properties.items():
                add_arguments_from_schema(
                    arguments_group,
                    child,
                    unparsed_arguments,
                    (*names[:-1], f'{names[-1]}[0]', name),
                )
        # If there aren't any children, then this is an array of scalars. Recurse accordingly.
        else:
            add_arguments_from_schema(
                arguments_group,
                items,
                unparsed_arguments,
                (*names[:-1], f'{names[-1]}[0]'),
            )

    flag_name = '.'.join(names).replace('_', '-')

    # Certain options already have corresponding flags on individual actions (like "create
    # --progress"), so don't bother adding them to the global flags.
    if not flag_name or flag_name in OMITTED_FLAG_NAMES:
        return

    metavar = names[-1].upper()
    description = make_argument_description(schema, flag_name)

    # The object=str and array=str given here is to support specifying an object or an array as a
    # YAML string on the command-line.
    argument_type = borgmatic.config.schema.parse_type(schema_type, object=str, array=str)

    # As a UX nicety, add separate true and false flags for boolean options.
    if schema_type == 'boolean':
        arguments_group.add_argument(
            f'--{flag_name}',
            action='store_true',
            default=None,
            help=description,
        )

        if names[-1].startswith('no_'):
            no_flag_name = '.'.join((*names[:-1], names[-1][len('no_') :])).replace('_', '-')
        else:
            no_flag_name = '.'.join((*names[:-1], 'no-' + names[-1])).replace('_', '-')

        arguments_group.add_argument(
            f'--{no_flag_name}',
            dest=flag_name.replace('-', '_'),
            action='store_false',
            default=None,
            help=f'Set the --{flag_name} value to false.',
        )
    elif flag_name == 'verbosity':
        arguments_group.add_argument(
            '-v',
            '--verbosity',
            type=argument_type,
            metavar=metavar,
            help=description,
        )
    else:
        arguments_group.add_argument(
            f'--{flag_name}',
            type=argument_type,
            metavar=metavar,
            help=description,
        )

    add_array_element_arguments(arguments_group, unparsed_arguments, flag_name)


def make_parsers(schema, unparsed_arguments):  # noqa: PLR0915
    '''
    Given a configuration schema dict and unparsed arguments as a sequence of strings, build a
    global arguments parser, individual action parsers, and a combined parser containing both.
    Return them as a tuple. The global parser is useful for parsing just global arguments while
    ignoring actions, and the combined parser is handy for displaying help that includes everything:
    global flags, a list of actions, etc.
    '''
    config_paths = collect.get_default_config_paths()

    # Using allow_abbrev=False here prevents the global parser from erroring about "ambiguous"
    # options like --encryption. Such options are intended for an action parser rather than the
    # global parser, and so we don't want to error on them here.
    global_parser = ArgumentParser(allow_abbrev=False, add_help=False)
    global_group = global_parser.add_argument_group('global arguments')

    global_group.add_argument(
        '-c',
        '--config',
        dest='config_paths',
        action='append',
        help='Configuration filename or directory, can specify flag multiple times, defaults to /etc/borgmatic/config.yaml, /etc/borgmatic.d, $XDG_CONFIG_HOME/borgmatic/config.yaml, and $XDG_CONFIG_HOME/borgmatic.d, where $XDG_CONFIG_HOME defaults to $HOME/.config',
    )
    global_group.add_argument(
        '-n',
        '--dry-run',
        dest='dry_run',
        action='store_true',
        help='Go through the motions, but do not actually write to any repositories',
    )
    global_group.add_argument(
        '--override',
        metavar='OPTION.SUBOPTION=VALUE',
        dest='overrides',
        action='append',
        help='Deprecated. Configuration file option to override with specified value, see documentation for overriding list or key/value options, can specify flag multiple times',
    )
    global_group.add_argument(
        '--no-environment-interpolation',
        dest='resolve_env',
        action='store_false',
        help='Do not resolve environment variables in configuration files',
    )
    global_group.add_argument(
        '--bash-completion',
        default=False,
        action='store_true',
        help='Show bash completion script and exit',
    )
    global_group.add_argument(
        '--fish-completion',
        default=False,
        action='store_true',
        help='Show fish completion script and exit',
    )
    global_group.add_argument(
        '--version',
        dest='version',
        default=False,
        action='store_true',
        help='Display installed version number of borgmatic and exit',
    )
    add_arguments_from_schema(global_group, schema, unparsed_arguments)

    global_plus_action_parser = ArgumentParser(
        description='''
            Simple, configuration-driven backup software for servers and workstations. If no actions
            are given, then borgmatic defaults to: create, prune, compact, and check.
            ''',
        parents=[global_parser],
    )

    action_parsers = global_plus_action_parser.add_subparsers(
        title='actions',
        metavar='',
        help='Specify zero or more actions. Defaults to create, prune, compact, and check. Use --help with action for details:',
    )
    repo_create_parser = action_parsers.add_parser(
        'repo-create',
        aliases=ACTION_ALIASES['repo-create'],
        help='Create a new, empty Borg repository (also known as "init")',
        description='Create a new, empty Borg repository (also known as "init")',
        add_help=False,
    )
    repo_create_group = repo_create_parser.add_argument_group('repo-create arguments')
    repo_create_group.add_argument(
        '-e',
        '--encryption',
        dest='encryption_mode',
        help='Borg repository encryption mode',
    )
    repo_create_group.add_argument(
        '--source-repository',
        '--other-repo',
        metavar='KEY_REPOSITORY',
        help='Path to an existing Borg repository whose key material should be reused [Borg 2.x+ only]',
    )
    repo_create_group.add_argument(
        '--repository',
        help='Path of the new repository to create (must be already specified in a borgmatic configuration file), defaults to the configured repository if there is only one, quoted globs supported',
    )
    repo_create_group.add_argument(
        '--copy-crypt-key',
        action='store_true',
        help='Copy the crypt key used for authenticated encryption from the source repository, defaults to a new random key [Borg 2.x+ only]',
    )
    repo_create_group.add_argument(
        '--append-only',
        default=None,
        action='store_true',
        help='Create an append-only repository',
    )
    repo_create_group.add_argument(
        '--storage-quota',
        help='Create a repository with a fixed storage quota',
    )
    repo_create_group.add_argument(
        '--make-parent-dirs',
        dest='make_parent_directories',
        default=None,
        action='store_true',
        help='Create any missing parent directories of the repository directory [Borg 1.x only]',
    )
    repo_create_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    transfer_parser = action_parsers.add_parser(
        'transfer',
        aliases=ACTION_ALIASES['transfer'],
        help='Transfer archives from one repository to another, optionally upgrading the transferred data [Borg 2.0+ only]',
        description='Transfer archives from one repository to another, optionally upgrading the transferred data [Borg 2.0+ only]',
        add_help=False,
    )
    transfer_group = transfer_parser.add_argument_group('transfer arguments')
    transfer_group.add_argument(
        '--repository',
        help='Path of existing destination repository to transfer archives to, defaults to the configured repository if there is only one, quoted globs supported',
    )
    transfer_group.add_argument(
        '--source-repository',
        help='Path of existing source repository to transfer archives from',
        required=True,
    )
    transfer_group.add_argument(
        '--archive',
        help='Name or hash of a single archive to transfer (or "latest"), defaults to transferring all archives',
    )
    transfer_group.add_argument(
        '--upgrader',
        help='Upgrader type used to convert the transferred data, e.g. "From12To20" to upgrade data from Borg 1.2 to 2.0 format, defaults to no conversion',
    )
    transfer_group.add_argument(
        '--progress',
        default=None,
        action='store_true',
        help='Display progress as each archive is transferred',
    )
    transfer_group.add_argument(
        '-a',
        '--match-archives',
        '--glob-archives',
        metavar='PATTERN',
        help='Only transfer archives with names, hashes, or series matching this pattern',
    )
    transfer_group.add_argument(
        '--sort-by',
        metavar='KEYS',
        help='Comma-separated list of sorting keys',
    )
    transfer_group.add_argument(
        '--first',
        metavar='N',
        help='Only transfer first N archives after other filters are applied',
    )
    transfer_group.add_argument(
        '--last',
        metavar='N',
        help='Only transfer last N archives after other filters are applied',
    )
    transfer_group.add_argument(
        '--oldest',
        metavar='TIMESPAN',
        help='Transfer archives within a specified time range starting from the timestamp of the oldest archive (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    transfer_group.add_argument(
        '--newest',
        metavar='TIMESPAN',
        help='Transfer archives within a time range that ends at timestamp of the newest archive and starts a specified time range ago (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    transfer_group.add_argument(
        '--older',
        metavar='TIMESPAN',
        help='Transfer archives that are older than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    transfer_group.add_argument(
        '--newer',
        metavar='TIMESPAN',
        help='Transfer archives that are newer than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    transfer_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    prune_parser = action_parsers.add_parser(
        'prune',
        aliases=ACTION_ALIASES['prune'],
        help='Prune archives according to the retention policy (with Borg 1.2+, you must run compact afterwards to actually free space)',
        description='Prune archives according to the retention policy (with Borg 1.2+, you must run compact afterwards to actually free space)',
        add_help=False,
    )
    prune_group = prune_parser.add_argument_group('prune arguments')
    prune_group.add_argument(
        '--repository',
        help='Path of specific existing repository to prune (must be already specified in a borgmatic configuration file), quoted globs supported',
    )
    prune_group.add_argument(
        '-a',
        '--match-archives',
        '--glob-archives',
        metavar='PATTERN',
        help='When pruning, only consider archives with names, hashes, or series matching this pattern',
    )
    prune_group.add_argument(
        '--stats',
        dest='statistics',
        default=None,
        action='store_true',
        help='Display statistics of the pruned archive [Borg 1 only]',
    )
    prune_group.add_argument(
        '--list',
        dest='list_details',
        default=None,
        action='store_true',
        help='List archives kept/pruned',
    )
    prune_group.add_argument(
        '--oldest',
        metavar='TIMESPAN',
        help='Prune archives within a specified time range starting from the timestamp of the oldest archive (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    prune_group.add_argument(
        '--newest',
        metavar='TIMESPAN',
        help='Prune archives within a time range that ends at timestamp of the newest archive and starts a specified time range ago (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    prune_group.add_argument(
        '--older',
        metavar='TIMESPAN',
        help='Prune archives that are older than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    prune_group.add_argument(
        '--newer',
        metavar='TIMESPAN',
        help='Prune archives that are newer than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    prune_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    compact_parser = action_parsers.add_parser(
        'compact',
        aliases=ACTION_ALIASES['compact'],
        help='Compact segments to free space [Borg 1.2+, borgmatic 1.5.23+ only]',
        description='Compact segments to free space [Borg 1.2+, borgmatic 1.5.23+ only]',
        add_help=False,
    )
    compact_group = compact_parser.add_argument_group('compact arguments')
    compact_group.add_argument(
        '--repository',
        help='Path of specific existing repository to compact (must be already specified in a borgmatic configuration file), quoted globs supported',
    )
    compact_group.add_argument(
        '--progress',
        default=None,
        action='store_true',
        help='Display progress as each segment is compacted',
    )
    compact_group.add_argument(
        '--cleanup-commits',
        dest='cleanup_commits',
        default=False,
        action='store_true',
        help='Cleanup commit-only 17-byte segment files left behind by Borg 1.1 [flag in Borg 1.2 only]',
    )
    compact_group.add_argument(
        '--threshold',
        type=int,
        dest='compact_threshold',
        help='Minimum saved space percentage threshold for compacting a segment, defaults to 10',
    )
    compact_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    create_parser = action_parsers.add_parser(
        'create',
        aliases=ACTION_ALIASES['create'],
        help='Create an archive (actually perform a backup)',
        description='Create an archive (actually perform a backup)',
        add_help=False,
    )
    create_group = create_parser.add_argument_group('create arguments')
    create_group.add_argument(
        '--repository',
        help='Path of specific existing repository to backup to (must be already specified in a borgmatic configuration file), quoted globs supported',
    )
    create_group.add_argument(
        '--progress',
        default=None,
        action='store_true',
        help='Display progress for each file as it is backed up',
    )
    create_group.add_argument(
        '--stats',
        dest='statistics',
        default=None,
        action='store_true',
        help='Display statistics of archive',
    )
    create_group.add_argument(
        '--list',
        '--files',
        dest='list_details',
        default=None,
        action='store_true',
        help='Show per-file details',
    )
    create_group.add_argument(
        '--json',
        dest='json',
        default=False,
        action='store_true',
        help='Output results as JSON',
    )
    create_group.add_argument(
        '--comment',
        metavar='COMMENT',
        help='Add a comment text to the archive',
    )
    create_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    check_parser = action_parsers.add_parser(
        'check',
        aliases=ACTION_ALIASES['check'],
        help='Check archives for consistency',
        description='Check archives for consistency',
        add_help=False,
    )
    check_group = check_parser.add_argument_group('check arguments')
    check_group.add_argument(
        '--repository',
        help='Path of specific existing repository to check (must be already specified in a borgmatic configuration file), quoted globs supported',
    )
    check_group.add_argument(
        '--progress',
        default=None,
        action='store_true',
        help='Display progress for each file as it is checked',
    )
    check_group.add_argument(
        '--repair',
        dest='repair',
        default=False,
        action='store_true',
        help='Attempt to repair any inconsistencies found (for interactive use)',
    )
    check_group.add_argument(
        '--max-duration',
        metavar='SECONDS',
        help='How long to check the repository before interrupting the check, defaults to no interruption',
    )
    check_group.add_argument(
        '-a',
        '--match-archives',
        '--glob-archives',
        metavar='PATTERN',
        help='Only check archives with names, hashes, or series matching this pattern',
    )
    check_group.add_argument(
        '--only',
        metavar='CHECK',
        choices=('repository', 'archives', 'data', 'extract', 'spot'),
        dest='only_checks',
        action='append',
        help='Run a particular consistency check (repository, archives, data, extract, or spot) instead of configured checks (subject to configured frequency, can specify flag multiple times)',
    )
    check_group.add_argument(
        '--force',
        default=False,
        action='store_true',
        help='Ignore configured check frequencies and run checks unconditionally',
    )
    check_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    delete_parser = action_parsers.add_parser(
        'delete',
        aliases=ACTION_ALIASES['delete'],
        help='Delete an archive from a repository or delete an entire repository (with Borg 1.2+, you must run compact afterwards to actually free space)',
        description='Delete an archive from a repository or delete an entire repository (with Borg 1.2+, you must run compact afterwards to actually free space)',
        add_help=False,
    )
    delete_group = delete_parser.add_argument_group('delete arguments')
    delete_group.add_argument(
        '--repository',
        help='Path of repository to delete or delete archives from, defaults to the configured repository if there is only one, quoted globs supported',
    )
    delete_group.add_argument(
        '--archive',
        help='Archive name, hash, or series to delete',
    )
    delete_group.add_argument(
        '--list',
        dest='list_details',
        default=None,
        action='store_true',
        help='Show details for the deleted archives',
    )
    delete_group.add_argument(
        '--stats',
        dest='statistics',
        default=None,
        action='store_true',
        help='Display statistics for the deleted archives',
    )
    delete_group.add_argument(
        '--cache-only',
        action='store_true',
        help='Delete only the local cache for the given repository',
    )
    delete_group.add_argument(
        '--force',
        action='count',
        help='Force deletion of corrupted archives, can be given twice if once does not work',
    )
    delete_group.add_argument(
        '--keep-security-info',
        action='store_true',
        help='Do not delete the local security info when deleting a repository',
    )
    delete_group.add_argument(
        '--save-space',
        action='store_true',
        help='Work slower, but using less space [Not supported in Borg 2.x+]',
    )
    delete_group.add_argument(
        '--checkpoint-interval',
        type=int,
        metavar='SECONDS',
        help='Write a checkpoint at the given interval, defaults to 1800 seconds (30 minutes)',
    )
    delete_group.add_argument(
        '-a',
        '--match-archives',
        '--glob-archives',
        metavar='PATTERN',
        help='Only delete archives with names, hashes, or series matching this pattern',
    )
    delete_group.add_argument(
        '--sort-by',
        metavar='KEYS',
        help='Comma-separated list of sorting keys',
    )
    delete_group.add_argument(
        '--first',
        metavar='N',
        help='Delete first N archives after other filters are applied',
    )
    delete_group.add_argument(
        '--last',
        metavar='N',
        help='Delete last N archives after other filters are applied',
    )
    delete_group.add_argument(
        '--oldest',
        metavar='TIMESPAN',
        help='Delete archives within a specified time range starting from the timestamp of the oldest archive (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    delete_group.add_argument(
        '--newest',
        metavar='TIMESPAN',
        help='Delete archives within a time range that ends at timestamp of the newest archive and starts a specified time range ago (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    delete_group.add_argument(
        '--older',
        metavar='TIMESPAN',
        help='Delete archives that are older than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    delete_group.add_argument(
        '--newer',
        metavar='TIMESPAN',
        help='Delete archives that are newer than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    delete_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    extract_parser = action_parsers.add_parser(
        'extract',
        aliases=ACTION_ALIASES['extract'],
        help='Extract files from a named archive to the current directory',
        description='Extract a named archive to the current directory',
        add_help=False,
    )
    extract_group = extract_parser.add_argument_group('extract arguments')
    extract_group.add_argument(
        '--repository',
        help='Path of repository to extract, defaults to the configured repository if there is only one, quoted globs supported',
    )
    extract_group.add_argument(
        '--archive',
        help='Name or hash of a single archive to extract (or "latest")',
        required=True,
    )
    extract_group.add_argument(
        '--path',
        '--restore-path',
        metavar='PATH',
        dest='paths',
        action='append',
        help='Path to extract from archive, can specify flag multiple times, defaults to the entire archive',
    )
    extract_group.add_argument(
        '--destination',
        metavar='PATH',
        dest='destination',
        help='Directory to extract files into, defaults to the current directory',
    )
    extract_group.add_argument(
        '--strip-components',
        type=lambda number: number if number == 'all' else int(number),
        metavar='NUMBER',
        help='Number of leading path components to remove from each extracted path or "all" to strip all leading path components. Skip paths with fewer elements',
    )
    extract_group.add_argument(
        '--progress',
        default=None,
        action='store_true',
        help='Display progress for each file as it is extracted',
    )
    extract_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    config_parser = action_parsers.add_parser(
        'config',
        aliases=ACTION_ALIASES['config'],
        help='Perform configuration file related operations',
        description='Perform configuration file related operations',
        add_help=False,
    )

    config_group = config_parser.add_argument_group('config arguments')
    config_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    config_parsers = config_parser.add_subparsers(
        title='config sub-actions',
    )

    config_bootstrap_parser = config_parsers.add_parser(
        'bootstrap',
        help='Extract the borgmatic configuration files from a named archive',
        description='Extract the borgmatic configuration files from a named archive',
        add_help=False,
    )
    config_bootstrap_group = config_bootstrap_parser.add_argument_group(
        'config bootstrap arguments',
    )
    config_bootstrap_group.add_argument(
        '--repository',
        help='Path of repository to extract config files from, quoted globs supported',
        required=True,
    )
    config_bootstrap_group.add_argument(
        '--local-path',
        help='Alternate Borg local executable. Defaults to "borg"',
        default='borg',
    )
    config_bootstrap_group.add_argument(
        '--remote-path',
        help='Alternate Borg remote executable. Defaults to "borg"',
        default='borg',
    )
    config_bootstrap_group.add_argument(
        '--user-runtime-directory',
        help='Path used for temporary runtime data like bootstrap metadata. Defaults to $XDG_RUNTIME_DIR or $TMPDIR or $TEMP or /var/run/$UID',
    )
    config_bootstrap_group.add_argument(
        '--borgmatic-source-directory',
        help='Deprecated. Path formerly used for temporary runtime data like bootstrap metadata. Defaults to ~/.borgmatic',
    )
    config_bootstrap_group.add_argument(
        '--archive',
        help='Name or hash of a single archive to extract config files from, defaults to "latest"',
        default='latest',
    )
    config_bootstrap_group.add_argument(
        '--destination',
        metavar='PATH',
        dest='destination',
        help='Directory to extract config files into, defaults to /',
        default='/',
    )
    config_bootstrap_group.add_argument(
        '--strip-components',
        type=lambda number: number if number == 'all' else int(number),
        metavar='NUMBER',
        help='Number of leading path components to remove from each extracted path or "all" to strip all leading path components. Skip paths with fewer elements',
    )
    config_bootstrap_group.add_argument(
        '--progress',
        default=None,
        action='store_true',
        help='Display progress for each file as it is extracted',
    )
    config_bootstrap_group.add_argument(
        '--ssh-command',
        metavar='COMMAND',
        help='Command to use instead of "ssh"',
    )
    config_bootstrap_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    config_generate_parser = config_parsers.add_parser(
        'generate',
        help='Generate a sample borgmatic configuration file',
        description='Generate a sample borgmatic configuration file',
        add_help=False,
    )
    config_generate_group = config_generate_parser.add_argument_group('config generate arguments')
    config_generate_group.add_argument(
        '-s',
        '--source',
        dest='source_filename',
        help='Optional configuration file to merge into the generated configuration, useful for upgrading your configuration',
    )
    config_generate_group.add_argument(
        '-d',
        '--destination',
        dest='destination_path',
        default=config_paths[0],
        help='Destination configuration file (or directory if using --split), default: /etc/borgmatic/config.yaml',
    )
    config_generate_group.add_argument(
        '--overwrite',
        default=False,
        action='store_true',
        help='Whether to overwrite any existing destination file, defaults to false',
    )
    config_generate_group.add_argument(
        '--split',
        action='store_true',
        help='Assuming the destination is a directory instead of a file, split the configuration into separate files within it, one per option, useful for documentation',
    )
    config_generate_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    config_validate_parser = config_parsers.add_parser(
        'validate',
        help='Validate borgmatic configuration files specified with --config (see borgmatic --help)',
        description='Validate borgmatic configuration files specified with --config (see borgmatic --help)',
        add_help=False,
    )
    config_validate_group = config_validate_parser.add_argument_group('config validate arguments')
    config_validate_group.add_argument(
        '-s',
        '--show',
        action='store_true',
        help='Show the validated configuration after all include merging has occurred',
    )
    config_validate_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    config_show_parser = config_parsers.add_parser(
        'show',
        help='Show the computed configuration for each file specified with --config (see borgmatic --help)',
        description='Show the computed configuration for each file specified with --config (see borgmatic --help)',
        add_help=False,
    )
    config_show_group = config_show_parser.add_argument_group('config show arguments')
    config_show_group.add_argument(
        '--option',
        help='Show the value of a single named configuration option instead of the entire configuration',
    )
    config_show_group.add_argument(
        '--json',
        action='store_true',
        help='Show the configuration as JSON with one array element per configuration file',
    )
    config_show_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    export_tar_parser = action_parsers.add_parser(
        'export-tar',
        aliases=ACTION_ALIASES['export-tar'],
        help='Export an archive to a tar-formatted file or stream',
        description='Export an archive to a tar-formatted file or stream',
        add_help=False,
    )
    export_tar_group = export_tar_parser.add_argument_group('export-tar arguments')
    export_tar_group.add_argument(
        '--repository',
        help='Path of repository to export from, defaults to the configured repository if there is only one, quoted globs supported',
    )
    export_tar_group.add_argument(
        '--archive',
        help='Name or hash of a single archive to export (or "latest")',
        required=True,
    )
    export_tar_group.add_argument(
        '--path',
        metavar='PATH',
        dest='paths',
        action='append',
        help='Path to export from archive, can specify flag multiple times, defaults to the entire archive',
    )
    export_tar_group.add_argument(
        '--destination',
        metavar='PATH',
        dest='destination',
        help='Path to destination export tar file, or "-" for stdout (but be careful about dirtying output with --verbosity or --list)',
        required=True,
    )
    export_tar_group.add_argument(
        '--tar-filter',
        help='Name of filter program to pipe data through',
    )
    export_tar_group.add_argument(
        '--list',
        '--files',
        dest='list_details',
        default=None,
        action='store_true',
        help='Show per-file details',
    )
    export_tar_group.add_argument(
        '--strip-components',
        type=int,
        metavar='NUMBER',
        dest='strip_components',
        help='Number of leading path components to remove from each exported path. Skip paths with fewer elements',
    )
    export_tar_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    mount_parser = action_parsers.add_parser(
        'mount',
        aliases=ACTION_ALIASES['mount'],
        help='Mount files from a named archive as a FUSE filesystem',
        description='Mount a named archive as a FUSE filesystem',
        add_help=False,
    )
    mount_group = mount_parser.add_argument_group('mount arguments')
    mount_group.add_argument(
        '--repository',
        help='Path of repository to use, defaults to the configured repository if there is only one, quoted globs supported',
    )
    mount_group.add_argument(
        '--archive',
        help='Name or hash of a single archive to mount (or "latest")',
    )
    mount_group.add_argument(
        '--mount-point',
        metavar='PATH',
        dest='mount_point',
        help='Path where filesystem is to be mounted',
        required=True,
    )
    mount_group.add_argument(
        '--path',
        metavar='PATH',
        dest='paths',
        action='append',
        help='Path to mount from archive, can specify multiple times, defaults to the entire archive',
    )
    mount_group.add_argument(
        '--foreground',
        dest='foreground',
        default=False,
        action='store_true',
        help='Stay in foreground until ctrl-C is pressed',
    )
    mount_group.add_argument(
        '--first',
        metavar='N',
        help='Mount first N archives after other filters are applied',
    )
    mount_group.add_argument(
        '--last',
        metavar='N',
        help='Mount last N archives after other filters are applied',
    )
    mount_group.add_argument(
        '--oldest',
        metavar='TIMESPAN',
        help='Mount archives within a specified time range starting from the timestamp of the oldest archive (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    mount_group.add_argument(
        '--newest',
        metavar='TIMESPAN',
        help='Mount archives within a time range that ends at timestamp of the newest archive and starts a specified time range ago (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    mount_group.add_argument(
        '--older',
        metavar='TIMESPAN',
        help='Mount archives that are older than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    mount_group.add_argument(
        '--newer',
        metavar='TIMESPAN',
        help='Mount archives that are newer than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    mount_group.add_argument('--options', dest='options', help='Extra Borg mount options')
    mount_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    umount_parser = action_parsers.add_parser(
        'umount',
        aliases=ACTION_ALIASES['umount'],
        help='Unmount a FUSE filesystem that was mounted with "borgmatic mount"',
        description='Unmount a mounted FUSE filesystem',
        add_help=False,
    )
    umount_group = umount_parser.add_argument_group('umount arguments')
    umount_group.add_argument(
        '--mount-point',
        metavar='PATH',
        dest='mount_point',
        help='Path of filesystem to unmount',
        required=True,
    )
    umount_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    repo_delete_parser = action_parsers.add_parser(
        'repo-delete',
        aliases=ACTION_ALIASES['repo-delete'],
        help='Delete an entire repository (with Borg 1.2+, you must run compact afterwards to actually free space)',
        description='Delete an entire repository (with Borg 1.2+, you must run compact afterwards to actually free space)',
        add_help=False,
    )
    repo_delete_group = repo_delete_parser.add_argument_group('delete arguments')
    repo_delete_group.add_argument(
        '--repository',
        help='Path of repository to delete, defaults to the configured repository if there is only one, quoted globs supported',
    )
    repo_delete_group.add_argument(
        '--list',
        dest='list_details',
        default=None,
        action='store_true',
        help='Show details for the archives in the given repository',
    )
    repo_delete_group.add_argument(
        '--force',
        action='count',
        help='Force deletion of corrupted archives, can be given twice if once does not work',
    )
    repo_delete_group.add_argument(
        '--cache-only',
        action='store_true',
        help='Delete only the local cache for the given repository',
    )
    repo_delete_group.add_argument(
        '--keep-security-info',
        action='store_true',
        help='Do not delete the local security info when deleting a repository',
    )
    repo_delete_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    restore_parser = action_parsers.add_parser(
        'restore',
        aliases=ACTION_ALIASES['restore'],
        help='Restore data source (e.g. database) dumps from a named archive',
        description='Restore data source (e.g. database) dumps from a named archive. (To extract files instead, use "borgmatic extract".)',
        add_help=False,
    )
    restore_group = restore_parser.add_argument_group('restore arguments')
    restore_group.add_argument(
        '--repository',
        help='Path of repository to restore from, defaults to the configured repository if there is only one, quoted globs supported',
    )
    restore_group.add_argument(
        '--archive',
        help='Name or hash of a single archive to restore from (or "latest")',
        required=True,
    )
    restore_group.add_argument(
        '--data-source',
        '--database',
        metavar='NAME',
        dest='data_sources',
        action='append',
        help="Name of data source (e.g. database) to restore from the archive, must be defined in borgmatic's configuration, can specify the flag multiple times, defaults to all data sources in the archive",
    )
    restore_group.add_argument(
        '--schema',
        metavar='NAME',
        dest='schemas',
        action='append',
        help='Name of schema to restore from the data source, can specify flag multiple times, defaults to all schemas. Schemas are only supported for PostgreSQL and MongoDB databases',
    )
    restore_group.add_argument(
        '--hostname',
        help='Database hostname to restore to. Defaults to the "restore_hostname" option in borgmatic\'s configuration',
    )
    restore_group.add_argument(
        '--port',
        help='Database port to restore to. Defaults to the "restore_port" option in borgmatic\'s configuration',
    )
    restore_group.add_argument(
        '--container',
        help='Container to restore to. Defaults to the "restore_container" option in borgmatic\'s configuration',
    )
    restore_group.add_argument(
        '--username',
        help='Username with which to connect to the database. Defaults to the "restore_username" option in borgmatic\'s configuration',
    )
    restore_group.add_argument(
        '--password',
        help='Password with which to connect to the restore database. Defaults to the "restore_password" option in borgmatic\'s configuration',
    )
    restore_group.add_argument(
        '--restore-path',
        help='Path to restore SQLite database dumps to. Defaults to the "restore_path" option in borgmatic\'s configuration',
    )
    restore_group.add_argument(
        '--original-label',
        help='The label where the dump to restore came from, only necessary if you need to disambiguate dumps',
    )
    restore_group.add_argument(
        '--original-hostname',
        help='The hostname where the dump to restore came from, only necessary if you need to disambiguate dumps',
    )
    restore_group.add_argument(
        '--original-container',
        help='The container where the dump to restore came from, only necessary if you need to disambiguate dumps',
    )
    restore_group.add_argument(
        '--original-port',
        type=int,
        help="The port where the dump to restore came from (if that port is in borgmatic's configuration), only necessary if you need to disambiguate dumps",
    )
    restore_group.add_argument(
        '--hook',
        help='The name of the data source hook for the dump to restore, only necessary if you need to disambiguate dumps',
    )
    restore_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    repo_list_parser = action_parsers.add_parser(
        'repo-list',
        aliases=ACTION_ALIASES['repo-list'],
        help='List repository',
        description='List the archives in a repository',
        add_help=False,
    )
    repo_list_group = repo_list_parser.add_argument_group('repo-list arguments')
    repo_list_group.add_argument(
        '--repository',
        help='Path of repository to list, defaults to the configured repositories, quoted globs supported',
    )
    repo_list_group.add_argument(
        '--short',
        default=False,
        action='store_true',
        help='Output only archive names',
    )
    repo_list_group.add_argument('--format', help='Borg format for the archive listing')
    repo_list_group.add_argument(
        '--json',
        default=False,
        action='store_true',
        help='Output results as JSON',
    )
    repo_list_group.add_argument(
        '-P',
        '--prefix',
        help='Deprecated. Only list archive names starting with this prefix',
    )
    repo_list_group.add_argument(
        '-a',
        '--match-archives',
        '--glob-archives',
        metavar='PATTERN',
        help='Only list archive names, hashes, or series matching this pattern',
    )
    repo_list_group.add_argument(
        '--sort-by',
        metavar='KEYS',
        help='Comma-separated list of sorting keys',
    )
    repo_list_group.add_argument(
        '--first',
        metavar='N',
        help='List first N archives after other filters are applied',
    )
    repo_list_group.add_argument(
        '--last',
        metavar='N',
        help='List last N archives after other filters are applied',
    )
    repo_list_group.add_argument(
        '--oldest',
        metavar='TIMESPAN',
        help='List archives within a specified time range starting from the timestamp of the oldest archive (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    repo_list_group.add_argument(
        '--newest',
        metavar='TIMESPAN',
        help='List archives within a time range that ends at timestamp of the newest archive and starts a specified time range ago (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    repo_list_group.add_argument(
        '--older',
        metavar='TIMESPAN',
        help='List archives that are older than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    repo_list_group.add_argument(
        '--newer',
        metavar='TIMESPAN',
        help='List archives that are newer than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    repo_list_group.add_argument(
        '--deleted',
        default=False,
        action='store_true',
        help="List only deleted archives that haven't yet been compacted [Borg 2.x+ only]",
    )
    repo_list_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    list_parser = action_parsers.add_parser(
        'list',
        aliases=ACTION_ALIASES['list'],
        help='List archive',
        description='List the files in an archive or search for a file across archives',
        add_help=False,
    )
    list_group = list_parser.add_argument_group('list arguments')
    list_group.add_argument(
        '--repository',
        help='Path of repository containing archive to list, defaults to the configured repositories, quoted globs supported',
    )
    list_group.add_argument(
        '--archive',
        help='Name or hash of a single archive to list (or "latest")',
    )
    list_group.add_argument(
        '--path',
        metavar='PATH',
        dest='paths',
        action='append',
        help='Path or pattern to list from a single selected archive (via "--archive"), can specify flag multiple times, defaults to listing the entire archive',
    )
    list_group.add_argument(
        '--find',
        metavar='PATH',
        dest='find_paths',
        action='append',
        help='Partial path or pattern to search for and list across multiple archives, can specify flag multiple times',
    )
    list_group.add_argument(
        '--short',
        default=False,
        action='store_true',
        help='Output only path names',
    )
    list_group.add_argument('--format', help='Borg format for the file listing')
    list_group.add_argument(
        '--json',
        default=False,
        action='store_true',
        help='Output results as JSON',
    )
    list_group.add_argument(
        '-P',
        '--prefix',
        help='Deprecated. Only list archive names starting with this prefix',
    )
    list_group.add_argument(
        '-a',
        '--match-archives',
        '--glob-archives',
        metavar='PATTERN',
        help='Only list archive names matching this pattern',
    )
    list_group.add_argument(
        '--sort-by',
        metavar='KEYS',
        help='Comma-separated list of sorting keys',
    )
    list_group.add_argument(
        '--first',
        metavar='N',
        help='List first N archives after other filters are applied',
    )
    list_group.add_argument(
        '--last',
        metavar='N',
        help='List last N archives after other filters are applied',
    )
    list_group.add_argument(
        '-e',
        '--exclude',
        metavar='PATTERN',
        help='Exclude paths matching the pattern',
    )
    list_group.add_argument(
        '--exclude-from',
        metavar='FILENAME',
        help='Exclude paths from exclude file, one per line',
    )
    list_group.add_argument('--pattern', help='Include or exclude paths matching a pattern')
    list_group.add_argument(
        '--patterns-from',
        metavar='FILENAME',
        help='Include or exclude paths matching patterns from pattern file, one per line',
    )
    list_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    repo_info_parser = action_parsers.add_parser(
        'repo-info',
        aliases=ACTION_ALIASES['repo-info'],
        help='Show repository summary information such as disk space used',
        description='Show repository summary information such as disk space used',
        add_help=False,
    )
    repo_info_group = repo_info_parser.add_argument_group('repo-info arguments')
    repo_info_group.add_argument(
        '--repository',
        help='Path of repository to show info for, defaults to the configured repository if there is only one, quoted globs supported',
    )
    repo_info_group.add_argument(
        '--json',
        dest='json',
        default=False,
        action='store_true',
        help='Output results as JSON',
    )
    repo_info_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    info_parser = action_parsers.add_parser(
        'info',
        aliases=ACTION_ALIASES['info'],
        help='Show archive summary information such as disk space used',
        description='Show archive summary information such as disk space used',
        add_help=False,
    )
    info_group = info_parser.add_argument_group('info arguments')
    info_group.add_argument(
        '--repository',
        help='Path of repository containing archive to show info for, defaults to the configured repository if there is only one, quoted globs supported',
    )
    info_group.add_argument(
        '--archive',
        help='Archive name, hash, or series to show info for (or "latest")',
    )
    info_group.add_argument(
        '--json',
        dest='json',
        default=False,
        action='store_true',
        help='Output results as JSON',
    )
    info_group.add_argument(
        '-P',
        '--prefix',
        help='Deprecated. Only show info for archive names starting with this prefix',
    )
    info_group.add_argument(
        '-a',
        '--match-archives',
        '--glob-archives',
        metavar='PATTERN',
        help='Only show info for archive names, hashes, or series matching this pattern',
    )
    info_group.add_argument(
        '--sort-by',
        metavar='KEYS',
        help='Comma-separated list of sorting keys',
    )
    info_group.add_argument(
        '--first',
        metavar='N',
        help='Show info for first N archives after other filters are applied',
    )
    info_group.add_argument(
        '--last',
        metavar='N',
        help='Show info for last N archives after other filters are applied',
    )
    info_group.add_argument(
        '--oldest',
        metavar='TIMESPAN',
        help='Show info for archives within a specified time range starting from the timestamp of the oldest archive (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    info_group.add_argument(
        '--newest',
        metavar='TIMESPAN',
        help='Show info for archives within a time range that ends at timestamp of the newest archive and starts a specified time range ago (e.g. 7d or 12m) [Borg 2.x+ only]',
    )
    info_group.add_argument(
        '--older',
        metavar='TIMESPAN',
        help='Show info for archives that are older than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    info_group.add_argument(
        '--newer',
        metavar='TIMESPAN',
        help='Show info for archives that are newer than the specified time range (e.g. 7d or 12m) from the current time [Borg 2.x+ only]',
    )
    info_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    break_lock_parser = action_parsers.add_parser(
        'break-lock',
        aliases=ACTION_ALIASES['break-lock'],
        help='Break the repository and cache locks left behind by Borg aborting',
        description='Break Borg repository and cache locks left behind by Borg aborting',
        add_help=False,
    )
    break_lock_group = break_lock_parser.add_argument_group('break-lock arguments')
    break_lock_group.add_argument(
        '--repository',
        help='Path of repository to break the lock for, defaults to the configured repository if there is only one, quoted globs supported',
    )
    break_lock_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    key_parser = action_parsers.add_parser(
        'key',
        aliases=ACTION_ALIASES['key'],
        help='Perform repository key related operations',
        description='Perform repository key related operations',
        add_help=False,
    )

    key_group = key_parser.add_argument_group('key arguments')
    key_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    key_parsers = key_parser.add_subparsers(
        title='key sub-actions',
    )

    key_export_parser = key_parsers.add_parser(
        'export',
        help='Export a copy of the repository key for safekeeping in case the original goes missing or gets damaged',
        description='Export a copy of the repository key for safekeeping in case the original goes missing or gets damaged',
        add_help=False,
    )
    key_export_group = key_export_parser.add_argument_group('key export arguments')
    key_export_group.add_argument(
        '--paper',
        action='store_true',
        help='Export the key in a text format suitable for printing and later manual entry',
    )
    key_export_group.add_argument(
        '--qr-html',
        action='store_true',
        help='Export the key in an HTML format suitable for printing and later manual entry or QR code scanning',
    )
    key_export_group.add_argument(
        '--repository',
        help='Path of repository to export the key for, defaults to the configured repository if there is only one, quoted globs supported',
    )
    key_export_group.add_argument(
        '--path',
        metavar='PATH',
        help='Path to export the key to, defaults to stdout (but be careful about dirtying the output with --verbosity)',
    )
    key_export_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    key_import_parser = key_parsers.add_parser(
        'import',
        help='Import a copy of the repository key from backup',
        description='Import a copy of the repository key from backup',
        add_help=False,
    )
    key_import_group = key_import_parser.add_argument_group('key import arguments')
    key_import_group.add_argument(
        '--paper',
        action='store_true',
        help='Import interactively from a backup done with --paper',
    )
    key_import_group.add_argument(
        '--repository',
        help='Path of repository to import the key from, defaults to the configured repository if there is only one, quoted globs supported',
    )
    key_import_group.add_argument(
        '--path',
        metavar='PATH',
        help='Path to import the key from backup, defaults to stdin',
    )
    key_import_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    key_change_passphrase_parser = key_parsers.add_parser(
        'change-passphrase',
        help='Change the passphrase protecting the repository key',
        description='Change the passphrase protecting the repository key',
        add_help=False,
    )
    key_change_passphrase_group = key_change_passphrase_parser.add_argument_group(
        'key change-passphrase arguments',
    )
    key_change_passphrase_group.add_argument(
        '--repository',
        help='Path of repository to change the passphrase for, defaults to the configured repository if there is only one, quoted globs supported',
    )
    key_change_passphrase_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    recreate_parser = action_parsers.add_parser(
        'recreate',
        aliases=ACTION_ALIASES['recreate'],
        help='Recreate an archive in a repository (with Borg 1.2+, you must run compact afterwards to actually free space)',
        description='Recreate an archive in a repository (with Borg 1.2+, you must run compact afterwards to actually free space)',
        add_help=False,
    )
    recreate_group = recreate_parser.add_argument_group('recreate arguments')
    recreate_group.add_argument(
        '--repository',
        help='Path of repository containing archive to recreate, defaults to the configured repository if there is only one, quoted globs supported',
    )
    recreate_group.add_argument(
        '--archive',
        help='Archive name, hash, or series to recreate, defaults to all archives in the repository (if specified), or all archives across all repositories',
    )
    recreate_group.add_argument(
        '--list',
        dest='list_details',
        default=None,
        action='store_true',
        help='Show per-file details',
    )
    recreate_group.add_argument(
        '--target',
        metavar='TARGET',
        help='Create a new archive from the specified archive (via --archive), without replacing it',
    )
    recreate_group.add_argument(
        '--comment',
        metavar='COMMENT',
        help='Add a comment text to the archive or, if an archive is not provided, to all matching archives',
    )
    recreate_group.add_argument(
        '--timestamp',
        metavar='TIMESTAMP',
        help='Manually override the archive creation date/time (UTC)',
    )
    recreate_group.add_argument(
        '-a',
        '--match-archives',
        '--glob-archives',
        dest='match_archives',
        metavar='PATTERN',
        help='Only consider archive names, hashes, or series matching this pattern [Borg 2.x+ only]',
    )
    recreate_group.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit',
    )

    diff_parser = action_parsers.add_parser(
        'diff',
        aliases=ACTION_ALIASES['diff'],
        help='This command finds differences (file contents, user/group/mode) between archives',
        description='This command finds differences (file contents, user/group/mode) between archives',
    )
    diff_group = diff_parser.add_argument_group('diff arguments')
    diff_group.add_argument(
        '--repository',
        help='Path of repository containing archive to diff, defaults to the configured repository if there is only one, quoted globs supported',
    )
    diff_group.add_argument(
        '--archive',
        help='Archive name, hash, or series to diff',
        required=True,
    )
    diff_group.add_argument(
        '--second-archive',
        help='Second archive name, hash, or series to diff',
        required=True,
    )
    diff_group.add_argument(
        '--same-chunker-params', action='store_true', help='Override check of chunker parameters'
    )
    diff_group.add_argument(
        '--sort-by',
        metavar='KEY',
        dest='sort_keys',
        action='append',
        help='Advanced sorting: specify field(s) to sort by. Prefix with > for descending or < for ascending (default)',
    )
    diff_group.add_argument(
        '--content-only',
        action='store_true',
        help='Only compare differences in content (exclude metadata differences)',
    )
    diff_group.add_argument(
        '--only-patterns',
        action='store_true',
        help='Run the diff according to borgmatic configured patterns (ie do not diff entire archives)',
    )
    borg_parser = action_parsers.add_parser(
        'borg',
        aliases=ACTION_ALIASES['borg'],
        help='Run an arbitrary Borg command',
        description="Run an arbitrary Borg command based on borgmatic's configuration",
        add_help=False,
    )
    borg_group = borg_parser.add_argument_group('borg arguments')
    borg_group.add_argument(
        '--repository',
        help='Path of repository to pass to Borg, defaults to the configured repositories, quoted globs supported',
    )
    borg_group.add_argument(
        '--archive',
        help='Archive name, hash, or series to pass to Borg (or "latest")',
    )
    borg_group.add_argument(
        '--',
        metavar='OPTION',
        dest='options',
        nargs='+',
        help='Options to pass to Borg, command first ("create", "list", etc). "--" is optional. To specify the repository or the archive, you must use --repository or --archive instead of providing them here.',
    )
    borg_group.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    return global_parser, action_parsers, global_plus_action_parser


HIGHLANDER_ACTION_ARGUMENTS_COUNT = 2  # 1 for "global" + 1 for the action


def parse_arguments(schema, *unparsed_arguments):
    '''
    Given a configuration schema dict and the command-line arguments with which this script was
    invoked and unparsed arguments as a sequence of strings, parse the arguments and return them as
    a dict mapping from action name (or "global") to an argparse.Namespace instance.

    Raise ValueError if the arguments cannot be parsed.
    Raise SystemExit with an error code of 0 if "--help" was requested.
    '''
    global_parser, action_parsers, global_plus_action_parser = make_parsers(
        schema,
        unparsed_arguments,
    )
    arguments, remaining_action_arguments = parse_arguments_for_actions(
        unparsed_arguments,
        action_parsers.choices,
        global_parser,
    )

    if not arguments['global'].config_paths:
        arguments['global'].config_paths = collect.get_default_config_paths()

    for action_name in ('bootstrap', 'generate', 'validate'):
        if action_name in arguments and len(arguments) > HIGHLANDER_ACTION_ARGUMENTS_COUNT:
            raise ValueError(
                f'The {action_name} action cannot be combined with other actions. Please run it separately.',
            )

    unknown_arguments = get_unparsable_arguments(remaining_action_arguments)

    if unknown_arguments:
        if '--help' in unknown_arguments or '-h' in unknown_arguments:
            global_plus_action_parser.print_help()
            sys.exit(0)

        global_plus_action_parser.print_usage()
        raise ValueError(
            f"Unrecognized argument{'s' if len(unknown_arguments) > 1 else ''}: {' '.join(unknown_arguments)}",
        )

    if (
        ('list' in arguments and 'repo-info' in arguments and arguments['list'].json)  # noqa: PLR0916
        or ('list' in arguments and 'info' in arguments and arguments['list'].json)
        or ('repo-info' in arguments and 'info' in arguments and arguments['repo-info'].json)
    ):
        raise ValueError('With the --json flag, multiple actions cannot be used together.')

    if 'list' in arguments and (arguments['list'].prefix and arguments['list'].match_archives):
        raise ValueError(
            'With the list action, only one of --prefix or --match-archives flags can be used.',
        )

    if 'repo-list' in arguments and (
        arguments['repo-list'].prefix and arguments['repo-list'].match_archives
    ):
        raise ValueError(
            'With the repo-list action, only one of --prefix or --match-archives flags can be used.',
        )

    if 'info' in arguments and (  # noqa: PLR0916
        (arguments['info'].archive and arguments['info'].prefix)
        or (arguments['info'].archive and arguments['info'].match_archives)
        or (arguments['info'].prefix and arguments['info'].match_archives)
    ):
        raise ValueError(
            'With the info action, only one of --archive, --prefix, or --match-archives flags can be used.',
        )

    if 'borg' in arguments and arguments['global'].dry_run:
        raise ValueError('With the borg action, --dry-run is not supported.')

    return arguments



import logging
import shlex

import borgmatic.borg.environment
import borgmatic.borg.feature
import borgmatic.config.paths
import borgmatic.execute
from borgmatic.borg import flags
from borgmatic.borg.pattern import write_patterns_file

logger = logging.getLogger(__name__)


def diff(
    repository,
    archive,
    second_archive,
    config,
    local_borg_version,
    diff_arguments,
    global_arguments,
    local_path,
    remote_path=None,
    patterns=None,
):
    '''
    Given a local or remote repository path, two archive names, a configuration dict, the local Borg
    version string, an argparse.Namespace of diff arguments, an argparse.Namespace of global
    arguments, optional local and remote Borg paths, executes the diff command with the given
    arguments.
    '''
    borgmatic.logger.add_custom_log_levels()

    lock_wait = config.get('lock_wait', None)
    exclude_flags = flags.make_exclude_flags(config)
    extra_borg_options = config.get('extra_borg_options', {}).get('diff', '')

    if diff_arguments.only_patterns:
        # Write patterns to a temporary file and use that file with --patterns-from.
        patterns_file = write_patterns_file(
            patterns,
            borgmatic.config.paths.get_working_directory(config),
        )
    else:
        patterns_file = None

    if borgmatic.borg.feature.available(
        borgmatic.borg.feature.Feature.NUMERIC_IDS, local_borg_version
    ):
        numeric_ids_flags = ('--numeric-ids',) if config.get('numeric_ids') else ()
    else:
        numeric_ids_flags = ('--numeric-owner',) if config.get('numeric_ids') else ()

    diff_command = (
        (local_path, 'diff')
        + (('--remote-path', remote_path) if remote_path else ())
        + ('--log-json',)
        + (('--lock-wait', str(lock_wait)) if lock_wait is not None else ())
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + (
            ('--patterns-from', patterns_file.name)
            if patterns_file and diff_arguments.only_patterns
            else ()
        )
        + exclude_flags
        + numeric_ids_flags
        + (('--same-chunker-params',) if diff_arguments.same_chunker_params else ())
        + (('--sort-by', ','.join(diff_arguments.sort_keys)) if diff_arguments.sort_keys else ())
        + (('--content-only',) if diff_arguments.content_only else ())
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + (
            (*flags.make_repository_flags(repository, local_borg_version), archive)
            if borgmatic.borg.feature.available(
                borgmatic.borg.feature.Feature.SEPARATE_REPOSITORY_ARCHIVE,
                local_borg_version,
            )
            else flags.make_repository_archive_flags(repository, archive, local_borg_version)
        )
        + (second_archive,)
    )

    borgmatic.execute.execute_command(
        full_command=diff_command,
        output_log_level=logging.ANSWER,
        environment=borgmatic.borg.environment.make_environment(config),
        working_directory=borgmatic.config.paths.get_working_directory(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )



import os
import pathlib

IS_A_HOOK = False


def get_contained_patterns(parent_directory, candidate_patterns):
    '''
    Given a parent directory and a set of candidate patterns potentially inside it, get the subset
    of contained patterns for which the parent directory is actually the parent, a grandparent, the
    very same directory, etc. The idea is if, say, "/var/log" and "/var/lib" are candidate pattern
    paths, but there's a parent directory (logical volume, dataset, subvolume, etc.) at "/var", then
    "/var" is what we want to snapshot.

    If a parent directory and a candidate pattern are on different devices, skip the pattern. That's
    because any snapshot of a parent directory won't actually include "contained" directories if
    they reside on separate devices.

    For this function to work, a candidate pattern path can't have any globs or other non-literal
    characters in the initial portion of the path that matches the parent directory. For instance, a
    parent directory of "/var" would match a candidate pattern path of "/var/log/*/data", but not a
    pattern path like "/v*/log/*/data".

    The one exception is that if a regular expression pattern path starts with "^", that will get
    stripped off for purposes of matching against a parent directory.

    As part of this, also mutate the given set of candidate patterns to remove any actually
    contained patterns from it. That way, this function can be called multiple times, successively
    processing candidate patterns until none are left—and avoiding assigning any candidate pattern
    to more than one parent directory.
    '''
    if not candidate_patterns:
        return ()

    parent_device = os.stat(parent_directory).st_dev if os.path.exists(parent_directory) else None

    contained_patterns = tuple(
        candidate
        for candidate in candidate_patterns
        for candidate_path in (pathlib.PurePath(candidate.path.lstrip('^')),)
        if (
            pathlib.PurePath(parent_directory) == candidate_path
            or pathlib.PurePath(parent_directory) in candidate_path.parents
        )
        if candidate.device == parent_device
    )
    candidate_patterns -= set(contained_patterns)

    return contained_patterns



import logging

import borgmatic.borg.extract
import borgmatic.borg.repo_list

logger = logging.getLogger(__name__)


def run_extract(
    config_filename,
    repository,
    config,
    local_borg_version,
    extract_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "extract" action for the given repository.
    '''
    logger.info(f'Extracting archive {extract_arguments.archive}')
    borgmatic.borg.extract.extract_archive(
        global_arguments.dry_run,
        repository['path'],
        borgmatic.borg.repo_list.resolve_archive_name(
            repository['path'],
            extract_arguments.archive,
            config,
            local_borg_version,
            global_arguments,
            local_path,
            remote_path,
        ),
        extract_arguments.paths,
        config,
        local_borg_version,
        global_arguments,
        local_path=local_path,
        remote_path=remote_path,
        destination_path=extract_arguments.destination,
        strip_components=extract_arguments.strip_components,
    )



import logging

import borgmatic.actions.json
import borgmatic.borg.repo_info

logger = logging.getLogger(__name__)


def run_repo_info(
    repository,
    config,
    local_borg_version,
    repo_info_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "repo-info" action for the given repository.

    If repo_info_arguments.json is True, yield the JSON output from the info for the repository.
    '''
    if not repo_info_arguments.json:
        logger.answer('Displaying repository summary information')

    json_output = borgmatic.borg.repo_info.display_repository_info(
        repository['path'],
        config,
        local_borg_version,
        repo_info_arguments=repo_info_arguments,
        global_arguments=global_arguments,
        local_path=local_path,
        remote_path=remote_path,
    )
    if json_output:
        yield borgmatic.actions.json.parse_json(json_output, repository.get('label'))



import logging

import borgmatic.logger

VERBOSITY_DISABLED = -2
VERBOSITY_ERROR = -1
VERBOSITY_ANSWER = 0
VERBOSITY_SOME = 1
VERBOSITY_LOTS = 2


def verbosity_to_log_level(verbosity):
    '''
    Given a borgmatic verbosity value, return the corresponding Python log level.
    '''
    borgmatic.logger.add_custom_log_levels()

    return {
        VERBOSITY_DISABLED: logging.DISABLED,
        VERBOSITY_ERROR: logging.ERROR,
        VERBOSITY_ANSWER: logging.ANSWER,
        VERBOSITY_SOME: logging.INFO,
        VERBOSITY_LOTS: logging.DEBUG,
    }.get(verbosity, logging.WARNING)


DEFAULT_VERBOSITIES = {
    'verbosity': 0,
    'syslog_verbosity': -2,
    'log_file_verbosity': 1,
    'monitoring_verbosity': 1,
}


def get_verbosity(configs, option_name):
    '''
    Given a dict from configuration filename to configuration dict, and the name of a configuration
    verbosity option, return the maximum verbosity value from that option across the given
    configuration files.
    '''
    try:
        return max(
            verbosity
            for config in configs.values()
            for verbosity in (config.get(option_name, DEFAULT_VERBOSITIES[option_name]),)
            if verbosity is not None
        )
    except ValueError:
        return DEFAULT_VERBOSITIES[option_name]



import logging

import requests

import borgmatic.hooks.credential.parse

logger = logging.getLogger(__name__)


TIMEOUT_SECONDS = 10


def initialize_monitor(
    ping_url,
    config,
    config_filename,
    monitoring_log_level,
    dry_run,
):  # pragma: no cover
    '''
    No initialization is necessary for this monitor.
    '''


def convert_string_to_array(value):
    value = '' if value is None else str(value)
    items = []

    for item in value.split(','):
        stripped = item.strip()

        if stripped:
            items.append(stripped)

    return items


PRIORITY_NAME_TO_ID = {
    'max': 5,
    'urgent': 5,
    'high': 4,
    'default': 3,
    'low': 2,
    'min': 1,
}


def ping_monitor(hook_config, config, config_filename, state, monitoring_log_level, dry_run):
    '''
    Ping the configured Ntfy topic. Use the given configuration filename in any log entries.
    If this is a dry run, then don't actually ping anything.
    '''
    run_states = hook_config.get('states', ['fail'])

    if state.name.lower() in run_states:
        dry_run_label = ' (dry run; not actually pinging)' if dry_run else ''
        default_priority = PRIORITY_NAME_TO_ID['default']
        state_config = hook_config.get(
            state.name.lower(),
            {
                'title': f'A borgmatic {state.name} event happened',
                'message': f'A borgmatic {state.name} event happened',
                'priority': default_priority,
                'tags': 'borgmatic',
            },
        )

        base_url = hook_config.get('server', 'https://ntfy.sh')
        topic = hook_config.get('topic')

        logger.info(f'Pinging ntfy topic {topic}{dry_run_label}')
        logger.debug(f'Using ntfy ping URL {base_url}')

        headers = {
            'User-Agent': 'borgmatic',
        }
        payload = {
            'topic': topic,
            'title': state_config.get('title'),
            'message': state_config.get('message'),
            'priority': PRIORITY_NAME_TO_ID.get(state_config.get('priority'), default_priority),
            'tags': convert_string_to_array(state_config.get('tags')),
        }

        try:
            username = borgmatic.hooks.credential.parse.resolve_credential(
                hook_config.get('username'),
                config,
            )
            password = borgmatic.hooks.credential.parse.resolve_credential(
                hook_config.get('password'),
                config,
            )
            access_token = borgmatic.hooks.credential.parse.resolve_credential(
                hook_config.get('access_token'),
                config,
            )
        except ValueError as error:
            logger.warning(f'Ntfy credential error: {error}')
            return

        auth = None

        if access_token is not None:
            if username or password:
                logger.warning(
                    'ntfy access_token is set but so is username/password, only using access_token',
                )

            headers['Authorization'] = f'Bearer {access_token}'
        elif (username and password) is not None:
            auth = requests.auth.HTTPBasicAuth(username, password)
            logger.info(f'Using basic auth with user {username} for ntfy')
        elif username is not None:
            logger.warning('Password missing for ntfy authentication, defaulting to no auth')
        elif password is not None:
            logger.warning('Username missing for ntfy authentication, defaulting to no auth')

        if not dry_run:
            logging.getLogger('urllib3').setLevel(logging.ERROR)
            try:
                response = requests.post(
                    base_url,
                    auth=auth,
                    timeout=TIMEOUT_SECONDS,
                    headers=headers,
                    json=payload,
                )
                if not response.ok:
                    response.raise_for_status()
            except requests.exceptions.RequestException as error:
                logger.warning(f'ntfy error: {error}')


def destroy_monitor(ping_url_or_uuid, config, monitoring_log_level, dry_run):  # pragma: no cover
    '''
    No destruction is necessary for this monitor.
    '''



def variants(flag_name):
    '''
    Given a flag name as a string, yield it and any variations that should be complete-able as well.
    For instance, for a string like "--foo[0].bar", yield "--foo[0].bar", "--foo[1].bar", ...,
    "--foo[9].bar".
    '''
    if '[0]' in flag_name:
        for index in range(10):
            yield flag_name.replace('[0]', f'[{index}]')

        return

    yield flag_name



import io
import logging

import ruamel.yaml

logger = logging.getLogger(__name__)


def set_values(config, keys, value):
    '''
    Given a hierarchy of configuration dicts, a sequence of parsed key strings, and a string value,
    descend into the hierarchy based on the keys to set the value into the right place.
    '''
    if not keys:
        return

    first_key = keys[0]
    if len(keys) == 1:
        if isinstance(config, list):
            raise ValueError(
                'When overriding a list option, the value must use list syntax (e.g., "[foo, bar]" or "[{key: value}]" as appropriate)',
            )

        config[first_key] = value
        return

    if first_key not in config:
        config[first_key] = {}

    set_values(config[first_key], keys[1:], value)


def convert_value_type(value, option_type):
    '''
    Given a string value and its schema type as a string, determine its logical type (string,
    boolean, integer, etc.), and return it converted to that type.

    If the option type is a string, leave the value as a string so that special characters in it
    don't get interpreted as YAML during conversion.

    Raise ruamel.yaml.error.YAMLError if there's a parse issue with the YAML.
    '''
    if option_type == 'string':
        return value

    return ruamel.yaml.YAML(typ='safe').load(io.StringIO(value))


LEGACY_SECTION_NAMES = {'location', 'storage', 'retention', 'consistency', 'output', 'hooks'}


def strip_section_names(parsed_override_key):
    '''
    Given a parsed override key as a tuple of option and suboption names, strip out any initial
    legacy section names, since configuration file normalization also strips them out.
    '''
    if parsed_override_key[0] in LEGACY_SECTION_NAMES:
        return parsed_override_key[1:]

    return parsed_override_key


def type_for_option(schema, option_keys):
    '''
    Given a configuration schema and a sequence of keys identifying an option, e.g.
    ('extra_borg_options', 'init'), return the schema type of that option as a string.

    Return None if the option or its type cannot be found in the schema.
    '''
    option_schema = schema

    try:
        for key in option_keys:
            option_schema = option_schema['properties'][key]
    except KeyError:
        return None

    try:
        return option_schema['type']
    except KeyError:
        return None


def parse_overrides(raw_overrides, schema):
    '''
    Given a sequence of configuration file override strings in the form of "option.suboption=value"
    and a configuration schema dict, parse and return a sequence of tuples (keys, values), where
    keys is a sequence of strings. For instance, given the following raw overrides:

        ['my_option.suboption=value1', 'other_option=value2']

    ... return this:

        (
            (('my_option', 'suboption'), 'value1'),
            (('other_option'), 'value2'),
        )

    Raise ValueError if an override can't be parsed.
    '''
    if not raw_overrides:
        return ()

    parsed_overrides = []

    try:
        for raw_override in raw_overrides:
            raw_keys, value = raw_override.split('=', 1)
            keys = tuple(raw_keys.split('.'))
            option_type = type_for_option(schema, keys)

            parsed_overrides.append(
                (
                    keys,
                    convert_value_type(value, option_type),
                ),
            )
    except ValueError:
        raise ValueError(
            f"Invalid override '{raw_override}'. Make sure you use the form: OPTION=VALUE or OPTION.SUBOPTION=VALUE",
        )
    except ruamel.yaml.error.YAMLError as error:
        raise ValueError(f"Invalid override '{raw_override}': {error.problem}")

    return tuple(parsed_overrides)


def apply_overrides(config, schema, raw_overrides):
    '''
    Given a configuration dict, a corresponding configuration schema dict, and a sequence of
    configuration file override strings in the form of "option.suboption=value", parse each override
    and set it into the configuration dict.

    Set the overrides into the configuration both with and without deprecated section names (if
    used), so that the overrides work regardless of whether the configuration is also using
    deprecated section names.
    '''
    overrides = parse_overrides(raw_overrides, schema)

    if overrides:
        logger.warning(
            "The --override flag is deprecated and will be removed from a future release. Instead, use a command-line flag corresponding to the configuration option you'd like to set.",
        )

    for keys, value in overrides:
        set_values(config, keys, value)
        set_values(config, strip_section_names(keys), value)



import collections
import contextlib
import io
import json
import os
import re

import ruamel.yaml

import borgmatic.config.schema
from borgmatic.config import load, normalize

INDENT = 4
SEQUENCE_INDENT = 2


def insert_newline_before_comment(config, field_name):
    '''
    Using some ruamel.yaml black magic, insert a blank line in the config right before the given
    field and its comments.
    '''
    config.ca.items[field_name][1].insert(
        0,
        ruamel.yaml.tokens.CommentToken('\n', ruamel.yaml.error.CommentMark(0), None),
    )


SCALAR_SCHEMA_TYPES = {'string', 'boolean', 'integer', 'number'}


def schema_to_sample_configuration(schema, source_config=None, level=0, parent_is_sequence=False):
    '''
    Given a loaded configuration schema and a source configuration, generate and return sample
    config for the schema. Include comments for each option based on the schema "description".

    If a source config is given, walk it alongside the given schema so that both can be taken into
    account when commenting out particular options in add_comments_to_configuration_object().
    '''
    schema_type = schema.get('type')
    example = schema.get('example')

    if borgmatic.config.schema.compare_types(schema_type, {'array'}):
        config = ruamel.yaml.comments.CommentedSeq(
            example
            if borgmatic.config.schema.compare_types(
                schema['items'].get('type'),
                SCALAR_SCHEMA_TYPES,
            )
            else [
                schema_to_sample_configuration(
                    schema['items'],
                    source_config,
                    level,
                    parent_is_sequence=True,
                ),
            ],
        )
        add_comments_to_configuration_sequence(config, schema, indent=(level * INDENT))
    elif borgmatic.config.schema.compare_types(schema_type, {'object'}):
        if source_config and isinstance(source_config, list) and isinstance(source_config[0], dict):
            source_config = source_config[0]

        config = (
            ruamel.yaml.comments.CommentedMap(
                [
                    (
                        field_name,
                        schema_to_sample_configuration(
                            sub_schema,
                            (source_config or {}).get(field_name, {}),
                            level + 1,
                        ),
                    )
                    for field_name, sub_schema in borgmatic.config.schema.get_properties(
                        schema,
                    ).items()
                ],
            )
            or example
        )
        indent = (level * INDENT) + (SEQUENCE_INDENT if parent_is_sequence else 0)
        add_comments_to_configuration_object(
            config,
            schema,
            source_config,
            indent=indent,
            skip_first_field=parent_is_sequence,
        )
    elif borgmatic.config.schema.compare_types(schema_type, SCALAR_SCHEMA_TYPES, match=all):
        return example
    else:
        raise ValueError(f'Schema at level {level} is unsupported: {schema}')

    return config


def comment_out_line(line):
    # If it's already is commented out (or empty), there's nothing further to do!
    stripped_line = line.lstrip()
    if not stripped_line or stripped_line.startswith('#'):
        return line

    # Comment out the names of optional options, inserting the '#' after any indent for aesthetics.
    matches = re.match(r'(\s*)', line)
    indent_spaces = matches.group(0) if matches else ''
    count_indent_spaces = len(indent_spaces)

    return '# '.join((indent_spaces, line[count_indent_spaces:]))


def transform_optional_configuration(rendered_config, comment_out=True):
    '''
    Post-process a rendered configuration string to comment out optional key/values, as determined
    by a sentinel in the comment before each key.

    The idea is that the pre-commented configuration prevents the user from having to comment out a
    bunch of configuration they don't care about to get to a minimal viable configuration file.

    Ideally ruamel.yaml would support commenting out keys during configuration generation, but it's
    not terribly easy to accomplish that way.

    If comment_out is False, then just strip the comment sentinel without actually commenting
    anything out.
    '''
    lines = []
    optional = False
    indent_characters = None
    indent_characters_at_sentinel = None

    for line in rendered_config.split('\n'):
        indent_characters = len(line) - len(line.lstrip())

        # Upon encountering an optional configuration option, comment out lines until the next blank
        # line.
        if line.strip().startswith(f'# {COMMENTED_OUT_SENTINEL}'):
            if comment_out is False:
                continue

            optional = True
            indent_characters_at_sentinel = indent_characters
            continue

        # Hit a blank line or dedented, so reset commenting.
        if not line.strip() or (
            indent_characters_at_sentinel is not None
            and indent_characters < indent_characters_at_sentinel
        ):
            optional = False
            indent_characters_at_sentinel = None

        lines.append(comment_out_line(line) if optional else line)

    return '\n'.join(lines)


RUAMEL_YAML_END_OF_DOCUMENT_MARKER = '...\n'


def render_configuration(config):
    '''
    Given a config data structure of nested OrderedDicts, render the config as YAML and return it.
    '''
    dumper = ruamel.yaml.YAML(typ='rt')
    dumper.indent(mapping=INDENT, sequence=INDENT + SEQUENCE_INDENT, offset=INDENT)
    rendered = io.StringIO()
    dumper.dump(
        config,
        rendered,
        # Dumping certain values (integers, for instance) causes ruamel.yaml to append an
        # end-of-document "..." marker. Strip it.
        transform=lambda dumped: dumped.removesuffix(RUAMEL_YAML_END_OF_DOCUMENT_MARKER),
    )

    return rendered.getvalue()


def write_configuration(config_filename, rendered_config, mode=0o600, overwrite=False):
    '''
    Given a target config filename and rendered config YAML, write it out to file. Create any
    containing directories as needed. But if the file already exists and overwrite is False,
    abort before writing anything.
    '''
    if not overwrite and os.path.exists(config_filename):
        raise FileExistsError(
            f'{config_filename} already exists. Aborting. Use --overwrite to replace the file.',
        )

    with contextlib.suppress(FileExistsError, FileNotFoundError):
        os.makedirs(os.path.dirname(config_filename), mode=0o700)

    with open(config_filename, 'w', encoding='utf-8') as config_file:
        config_file.write(rendered_config)

    os.chmod(config_filename, mode)


def add_comments_to_configuration_sequence(config, schema, indent=0):
    '''
    If the given config sequence's items are object, then mine the schema for the description of the
    object's first item, and slap that atop the sequence. Indent the comment the given number of
    characters.

    Doing this for sequences of maps results in nice comments that look like:

    ```
    things:
        # First key description. Added by this function.
        - key: foo
          # Second key description. Added by add_comments_to_configuration_object().
          other: bar
    ```
    '''
    if schema['items'].get('type') != 'object':
        return

    for field_name in config[0]:
        field_schema = borgmatic.config.schema.get_properties(schema['items']).get(field_name, {})
        description = field_schema.get('description')

        # No description to use? Skip it.
        if not field_schema or not description:
            return

        config[0].yaml_set_start_comment(description, indent=indent)

        # We only want the first key's description here, as the rest of the keys get commented by
        # add_comments_to_configuration_object().
        return


DEFAULT_KEYS = {'source_directories', 'repositories', 'keep_daily'}
COMMENTED_OUT_SENTINEL = 'COMMENT_OUT'


def add_comments_to_configuration_object(
    config,
    schema,
    source_config=None,
    indent=0,
    skip_first_field=False,
):
    '''
    Using descriptions from a schema as a source, add those descriptions as comments to the given
    configuration dict, putting them before each field. Indent the comment the given number of
    characters.

    If skip_first_field is True, omit the comment for the initial field. This is useful for
    sequences, where the comment for the first field goes before the sequence itself.

    And a sentinel for commenting out options that are neither in DEFAULT_KEYS nor the the given
    source configuration dict. The idea is that any options used in the source configuration should
    stay active in the generated configuration.
    '''
    for index, field_name in enumerate(config.keys()):
        if skip_first_field and index == 0:
            continue

        field_schema = borgmatic.config.schema.get_properties(schema).get(field_name, {})
        description = field_schema.get('description', '').strip()

        # If this isn't a default key, add an indicator to the comment, flagging it to be commented
        # out from the sample configuration. This sentinel is consumed by downstream processing that
        # does the actual commenting out.
        if field_name not in DEFAULT_KEYS and (
            source_config is None or field_name not in source_config
        ):
            description = (
                f'{description}\n{COMMENTED_OUT_SENTINEL}'
                if description
                else COMMENTED_OUT_SENTINEL
            )

        # No description to use? Skip it.
        if not field_schema or not description:  # pragma: no cover
            continue

        config.yaml_set_comment_before_after_key(key=field_name, before=description, indent=indent)

        if index > 0:
            insert_newline_before_comment(config, field_name)


RUAMEL_YAML_COMMENTS_INDEX = 1


def merge_source_configuration_into_destination(destination_config, source_config):
    '''
    Deep merge the given source configuration dict into the destination configuration CommentedMap,
    favoring values from the source when there are collisions.

    The purpose of this is to upgrade configuration files from old versions of borgmatic by adding
    new configuration keys and comments.
    '''
    if not source_config:
        return destination_config

    if not destination_config or not isinstance(source_config, collections.abc.Mapping):
        return source_config

    for field_name, source_value in source_config.items():
        # This is a mapping. Recurse for this key/value.
        if isinstance(source_value, collections.abc.Mapping):
            destination_config[field_name] = merge_source_configuration_into_destination(
                destination_config[field_name],
                source_value,
            )
            continue

        # This is a sequence. Recurse for each item in it.
        if isinstance(source_value, collections.abc.Sequence) and not isinstance(source_value, str):
            destination_value = destination_config[field_name]
            destination_config[field_name] = ruamel.yaml.comments.CommentedSeq(
                [
                    merge_source_configuration_into_destination(
                        destination_value[index] if index < len(destination_value) else None,
                        source_item,
                    )
                    for index, source_item in enumerate(source_value)
                ],
            )
            continue

        # This is some sort of scalar. Set it into the destination.
        destination_config[field_name] = source_value

    return destination_config


def get_configuration_subset(config, option_name):  # pragma: no cover
    '''
    Given configuration as a ruamel.yaml.CommentedMap and an option name found within it at the top
    level, return a new CommentedMap containing a subset of the configuration with only the given
    option and no other top-level options.

    This is useful when generating the sample configuration for a single option instead of a whole
    configuration file.
    '''
    option_config = ruamel.yaml.CommentedMap({option_name: config[option_name]})

    # Due to a quirk of ruamel.yaml, the comment right before a top-level key is not on that key and
    # needs to get copied separately.
    option_config.ca.items[option_name] = config.ca.items[option_name]

    return option_config


def generate_sample_configuration(
    dry_run,
    source_filename,
    destination_path,
    schema_filename,
    overwrite=False,
    split=False,
):
    '''
    Given an optional source configuration filename, a required destination configuration path, the
    path to a schema filename in a YAML rendition of the JSON Schema format, whether to overwrite a
    destination file, and whether to split the configuration into multiple files (one per option) in
    the assumed destination directory, write out sample configuration file(s) based on that schema.
    If a source filename is provided, merge the parsed contents of that configuration into the
    generated configuration.
    '''
    schema = ruamel.yaml.YAML(typ='safe').load(open(schema_filename, encoding='utf-8'))
    source_config = None

    if source_filename:
        source_config = load.load_configuration(source_filename)
        normalize.normalize(source_filename, source_config)

        # The borgmatic.config.normalize.normalize() function tacks on an empty "bootstrap" if
        # needed, so the hook gets used by default. But we don't want it to end up in the generated
        # config unless the user has set it explicitly, as an empty "bootstrap:" won't validate.
        if source_config and source_config.get('bootstrap') == {}:
            del source_config['bootstrap']

    destination_config = merge_source_configuration_into_destination(
        schema_to_sample_configuration(schema, source_config),
        source_config,
    )

    if dry_run:
        return

    if split:
        if os.path.exists(destination_path) and not os.path.isdir(destination_path):
            raise ValueError('With the --split flag, the destination path must be a directory')

        os.makedirs(destination_path, exist_ok=True)

        for option_name in destination_config:
            write_configuration(
                os.path.join(destination_path, f'{option_name}.yaml'),
                transform_optional_configuration(
                    render_configuration(get_configuration_subset(destination_config, option_name)),
                    comment_out=False,
                ).strip(),
                overwrite=overwrite,
            )

        # Also dump a manifest listing all the options we've written.
        json.dump(
            {'option_names': list(destination_config.keys())},
            open(os.path.join(destination_path, 'options.json'), 'w', encoding='utf-8'),
        )

        return

    if os.path.exists(destination_path) and not os.path.isfile(destination_path):
        raise ValueError('Without the --split flag, the destination path must be a file')

    write_configuration(
        destination_path,
        transform_optional_configuration(
            render_configuration(destination_config), comment_out=True
        ),
        overwrite=overwrite,
    )



import logging

import borgmatic.borg.break_lock

logger = logging.getLogger(__name__)


def run_break_lock(
    repository,
    config,
    local_borg_version,
    break_lock_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "break-lock" action for the given repository.
    '''
    logger.info('Breaking repository and cache locks')
    borgmatic.borg.break_lock.break_lock(
        repository['path'],
        config,
        local_borg_version,
        global_arguments,
        local_path=local_path,
        remote_path=remote_path,
    )



import logging
import shlex

import borgmatic.config.paths
import borgmatic.execute
import borgmatic.logger
from borgmatic.borg import environment, flags

logger = logging.getLogger(__name__)


def change_passphrase(
    repository_path,
    config,
    local_borg_version,
    change_passphrase_arguments,
    global_arguments,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a local or remote repository path, a configuration dict, the local Borg version, change
    passphrase arguments, and optional local and remote Borg paths, change the repository passphrase
    based on an interactive prompt.
    '''
    borgmatic.logger.add_custom_log_levels()
    umask = config.get('umask', None)
    lock_wait = config.get('lock_wait', None)
    extra_borg_options = config.get('extra_borg_options', {}).get('key_change_passphrase', '')

    full_command = (
        (local_path, 'key', 'change-passphrase')
        + (('--remote-path', remote_path) if remote_path else ())
        + (('--umask', str(umask)) if umask else ())
        + (('--lock-wait', str(lock_wait)) if lock_wait else ())
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + flags.make_repository_flags(
            repository_path,
            local_borg_version,
        )
    )

    if global_arguments.dry_run:
        logger.info('Skipping change password (dry run)')
        return

    # If the original passphrase is set programmatically, then Borg won't prompt for a new one! So
    # don't give Borg any passphrase, and it'll ask the user for both old and new ones.
    config_without_passphrase = {
        option_name: value
        for (option_name, value) in config.items()
        if option_name not in {'encryption_passphrase', 'encryption_passcommand'}
    }

    borgmatic.execute.execute_command(
        full_command,
        output_file=borgmatic.execute.DO_NOT_CAPTURE,
        output_log_level=logging.ANSWER,
        environment=environment.make_environment(config_without_passphrase),
        working_directory=borgmatic.config.paths.get_working_directory(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )

    logger.answer(
        "Don't forget to update your encryption_passphrase option (if needed)",
    )



import argparse
import json
import logging
import shlex

import borgmatic.config.paths
import borgmatic.logger
from borgmatic.borg import environment, feature, flags
from borgmatic.execute import execute_command_and_capture_output

logger = logging.getLogger(__name__)


def resolve_archive_name(
    repository_path,
    archive,
    config,
    local_borg_version,
    global_arguments,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a local or remote repository path, an archive name, a configuration dict, the local Borg
    version, global arguments as an argparse.Namespace, a local Borg path, and a remote Borg path,
    return the archive name. But if the archive name is "latest", then instead introspect the
    repository for the latest archive and return its name or ID, depending on whether the version of
    Borg in use supports archive series—different archives that share the same name but have unique
    IDs.

    Raise ValueError if "latest" is given but there are no archives in the repository.
    '''
    if archive != 'latest':
        return archive

    latest_archive = get_latest_archive(
        repository_path,
        config,
        local_borg_version,
        global_arguments,
        local_path=local_path,
        remote_path=remote_path,
    )

    return (
        latest_archive['id']
        if feature.available(feature.Feature.ARCHIVE_SERIES, local_borg_version)
        else latest_archive['name']
    )


def get_latest_archive(
    repository_path,
    config,
    local_borg_version,
    global_arguments,
    local_path='borg',
    remote_path=None,
    consider_checkpoints=False,
):
    '''
    Returns a dict with information about the latest archive of a repository.

    Raises ValueError if there are no archives in the repository.
    '''
    extra_borg_options = config.get('extra_borg_options', {}).get(
        'repo_list' if feature.available(feature.Feature.REPO_LIST, local_borg_version) else 'list',
        '',
    )

    full_command = (
        local_path,
        (
            'repo-list'
            if feature.available(feature.Feature.REPO_LIST, local_borg_version)
            else 'list'
        ),
        *flags.make_flags('remote-path', remote_path),
        *flags.make_flags('umask', config.get('umask')),
        *('--log-json',),
        *flags.make_flags('lock-wait', config.get('lock_wait')),
        *(
            flags.make_match_archives_flags(
                config.get('match_archives'),
                config.get('archive_name_format'),
                local_borg_version,
            )
        ),
        *(
            flags.make_flags('consider-checkpoints', consider_checkpoints)
            if not feature.available(feature.Feature.REPO_LIST, local_borg_version)
            else ()
        ),
        *flags.make_flags('last', 1),
        '--json',
        *(tuple(shlex.split(extra_borg_options)) if extra_borg_options else ()),
        *flags.make_repository_flags(repository_path, local_borg_version),
    )

    json_output = '\n'.join(
        execute_command_and_capture_output(
            full_command,
            environment=environment.make_environment(config),
            working_directory=borgmatic.config.paths.get_working_directory(config),
            borg_local_path=local_path,
            borg_exit_codes=config.get('borg_exit_codes'),
        )
    )

    archives = json.loads(json_output)['archives']

    try:
        latest_archive = archives[-1]
    except IndexError:
        raise ValueError('No archives found in the repository')

    logger.debug(f'Latest archive is {latest_archive["name"]} ({latest_archive["id"]})')

    return latest_archive


MAKE_FLAGS_EXCLUDES = ('repository', 'format', 'prefix', 'match_archives')


def make_repo_list_command(
    repository_path,
    config,
    local_borg_version,
    repo_list_arguments,
    global_arguments,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a local or remote repository path, a configuration dict, the local Borg version, the
    arguments to the repo_list action, global arguments as an argparse.Namespace instance, and local and
    remote Borg paths, return a command as a tuple to list archives with a repository.
    '''
    extra_borg_options = config.get('extra_borg_options', {}).get(
        'repo_list' if feature.available(feature.Feature.REPO_LIST, local_borg_version) else 'list',
        '',
    )

    return (
        (
            local_path,
            (
                'repo-list'
                if feature.available(feature.Feature.REPO_LIST, local_borg_version)
                else 'list'
            ),
        )
        + (
            ('--info',)
            if logger.getEffectiveLevel() == logging.INFO and not repo_list_arguments.json
            else ()
        )
        + (
            ('--debug', '--show-rc')
            if logger.isEnabledFor(logging.DEBUG) and not repo_list_arguments.json
            else ()
        )
        + flags.make_flags('remote-path', remote_path)
        + flags.make_flags('umask', config.get('umask'))
        + ('--log-json',)
        + flags.make_flags('lock-wait', config.get('lock_wait'))
        + (
            (
                flags.make_flags('match-archives', f'sh:{repo_list_arguments.prefix}*')
                if feature.available(feature.Feature.MATCH_ARCHIVES, local_borg_version)
                else flags.make_flags('glob-archives', f'{repo_list_arguments.prefix}*')
            )
            if repo_list_arguments.prefix
            else (
                flags.make_match_archives_flags(
                    config.get('match_archives'),
                    config.get('archive_name_format'),
                    local_borg_version,
                )
            )
        )
        + flags.make_flags(
            'format', repo_list_arguments.format or config.get('archive_list_format')
        )
        + flags.make_flags_from_arguments(repo_list_arguments, excludes=MAKE_FLAGS_EXCLUDES)
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + flags.make_repository_flags(repository_path, local_borg_version)
    )


def list_repository(
    repository_path,
    config,
    local_borg_version,
    repo_list_arguments,
    global_arguments,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a local or remote repository path, a configuration dict, the local Borg version, the
    arguments to the list action, global arguments as an argparse.Namespace instance, and local and
    remote Borg paths, display the output of listing Borg archives in the given repository (or
    return JSON output).
    '''
    borgmatic.logger.add_custom_log_levels()

    main_command = make_repo_list_command(
        repository_path,
        config,
        local_borg_version,
        repo_list_arguments,
        global_arguments,
        local_path,
        remote_path,
    )
    json_command = make_repo_list_command(
        repository_path,
        config,
        local_borg_version,
        argparse.Namespace(**dict(repo_list_arguments.__dict__, json=True)),
        global_arguments,
        local_path,
        remote_path,
    )
    working_directory = borgmatic.config.paths.get_working_directory(config)
    borg_exit_codes = config.get('borg_exit_codes')

    if repo_list_arguments.json:
        return '\n'.join(
            execute_command_and_capture_output(
                json_command,
                environment=environment.make_environment(config),
                working_directory=working_directory,
                borg_local_path=local_path,
                borg_exit_codes=borg_exit_codes,
            )
        )

    output_lines = tuple(
        execute_command_and_capture_output(
            main_command,
            output_log_level=logging.ANSWER,
            environment=environment.make_environment(config),
            working_directory=working_directory,
            borg_local_path=local_path,
            borg_exit_codes=borg_exit_codes,
        )
    )

    flags.warn_for_aggressive_archive_flags(main_command, output_lines)

    return None



from enum import Enum

from packaging.version import parse


class Feature(Enum):
    COMPACT = 1
    ATIME = 2
    NOFLAGS = 3
    NUMERIC_IDS = 4
    UPLOAD_RATELIMIT = 5
    SEPARATE_REPOSITORY_ARCHIVE = 6
    REPO_CREATE = 7
    REPO_LIST = 8
    REPO_INFO = 9
    REPO_DELETE = 10
    MATCH_ARCHIVES = 11
    EXCLUDED_FILES_MINUS = 12
    ARCHIVE_SERIES = 13
    NO_PRUNE_STATS = 14
    DRY_RUN_COMPACT = 15


FEATURE_TO_MINIMUM_BORG_VERSION = {
    Feature.COMPACT: parse('1.2.0a2'),  # borg compact
    Feature.ATIME: parse('1.2.0a7'),  # borg create --atime
    Feature.NOFLAGS: parse('1.2.0a8'),  # borg create --noflags
    Feature.NUMERIC_IDS: parse('1.2.0b3'),  # borg create/extract/mount --numeric-ids
    Feature.UPLOAD_RATELIMIT: parse('1.2.0b3'),  # borg create --upload-ratelimit
    Feature.SEPARATE_REPOSITORY_ARCHIVE: parse('2.0.0a2'),  # --repo with separate archive
    Feature.REPO_CREATE: parse('2.0.0a2'),  # borg repo-create
    Feature.REPO_LIST: parse('2.0.0a2'),  # borg repo-list
    Feature.REPO_INFO: parse('2.0.0a2'),  # borg repo-info
    Feature.REPO_DELETE: parse('2.0.0a2'),  # borg repo-delete
    Feature.MATCH_ARCHIVES: parse('2.0.0b3'),  # borg --match-archives
    Feature.EXCLUDED_FILES_MINUS: parse('2.0.0b5'),  # --list --filter uses "-" for excludes
    Feature.ARCHIVE_SERIES: parse('2.0.0b11'),  # identically named archives form a series
    Feature.NO_PRUNE_STATS: parse('2.0.0b10'),  # prune --stats is not available
    Feature.DRY_RUN_COMPACT: parse('1.2.9'),  # borg compact --dry-run support
}


def available(feature, borg_version):
    '''
    Given a Borg Feature constant and a Borg version string, return whether that feature is
    available in that version of Borg.
    '''
    return FEATURE_TO_MINIMUM_BORG_VERSION[feature] <= parse(borg_version)



import calendar
import contextlib
import datetime
import hashlib
import itertools
import logging
import os
import pathlib
import random
import shlex
import shutil
import subprocess
import textwrap

import borgmatic.actions.config.bootstrap
import borgmatic.actions.pattern
import borgmatic.borg.check
import borgmatic.borg.create
import borgmatic.borg.environment
import borgmatic.borg.extract
import borgmatic.borg.list
import borgmatic.borg.pattern
import borgmatic.borg.repo_list
import borgmatic.config.paths
import borgmatic.execute

DEFAULT_CHECKS = (
    {'name': 'repository', 'frequency': '1 month'},
    {'name': 'archives', 'frequency': '1 month'},
)


logger = logging.getLogger(__name__)


def parse_checks(config, only_checks=None):
    '''
    Given a configuration dict with a "checks" sequence of dicts and an optional list of override
    checks, return a tuple of named checks to run.

    For example, given a config of:

        {'checks': ({'name': 'repository'}, {'name': 'archives'})}

    This will be returned as:

        ('repository', 'archives')

    If no "checks" option is present in the config, return the DEFAULT_CHECKS. If a checks value
    has a name of "disabled", return an empty tuple, meaning that no checks should be run.
    '''
    checks = only_checks or tuple(
        check_config['name'] for check_config in (config.get('checks', None) or DEFAULT_CHECKS)
    )
    checks = tuple(check.lower() for check in checks)

    if 'disabled' in checks:
        logger.warning(
            'The "disabled" value for the "checks" option is deprecated and will be removed from a future release; use "skip_actions" instead',
        )

        if len(checks) > 1:
            logger.warning(
                'Multiple checks are configured, but one of them is "disabled"; not running any checks',
            )

        return ()

    return checks


def parse_frequency(frequency):
    '''
    Given a frequency string with a number and a unit of time, return a corresponding
    datetime.timedelta instance or None if the frequency is None or "always".

    For instance, given "3 weeks", return datetime.timedelta(weeks=3)

    Raise ValueError if the given frequency cannot be parsed.
    '''
    if not frequency:
        return None

    frequency = frequency.strip().lower()

    if frequency == 'always':
        return None

    try:
        number, time_unit = frequency.split(' ')
        number = int(number)
    except ValueError:
        raise ValueError(f"Could not parse consistency check frequency '{frequency}'")

    if not time_unit.endswith('s'):
        time_unit += 's'

    if time_unit == 'months':
        number *= 30
        time_unit = 'days'
    elif time_unit == 'years':
        number *= 365
        time_unit = 'days'

    try:
        return datetime.timedelta(**{time_unit: number})
    except TypeError:
        raise ValueError(f"Could not parse consistency check frequency '{frequency}'")


WEEKDAY_DAYS = calendar.day_name[0:5]
WEEKEND_DAYS = calendar.day_name[5:7]


def filter_checks_on_frequency(
    config,
    borg_repository_id,
    checks,
    force,
    archives_check_id=None,
    datetime_now=datetime.datetime.now,
):
    '''
    Given a configuration dict with a "checks" sequence of dicts, a Borg repository ID, a sequence
    of checks, whether to force checks to run, and an ID for the archives check potentially being
    run (if any), filter down those checks based on the configured "frequency" for each check as
    compared to its check time file.

    In other words, a check whose check time file's timestamp is too new (based on the configured
    frequency) will get cut from the returned sequence of checks. Example:

    config = {
        'checks': [
            {
                'name': 'archives',
                'frequency': '2 weeks',
            },
        ]
    }

    When this function is called with that config and "archives" in checks, "archives" will get
    filtered out of the returned result if its check time file is newer than 2 weeks old, indicating
    that it's not yet time to run that check again.

    Raise ValueError if a frequency cannot be parsed.
    '''
    if not checks:
        return checks

    filtered_checks = list(checks)

    if force:
        return tuple(filtered_checks)

    for check_config in config.get('checks', DEFAULT_CHECKS):
        check = check_config['name']
        if checks and check not in checks:
            continue

        only_run_on = check_config.get('only_run_on')
        if only_run_on:
            # Use a dict instead of a set to preserve ordering.
            days = dict.fromkeys(only_run_on)

            if 'weekday' in days:
                days = {
                    **dict.fromkeys(day for day in days if day != 'weekday'),
                    **dict.fromkeys(WEEKDAY_DAYS),
                }

            if 'weekend' in days:
                days = {
                    **dict.fromkeys(day for day in days if day != 'weekend'),
                    **dict.fromkeys(WEEKEND_DAYS),
                }

            if calendar.day_name[datetime_now().weekday()] not in days:
                logger.info(
                    f"Skipping {check} check due to day of the week; check only runs on {'/'.join(day.title() for day in days)} (use --force to check anyway)",
                )
                filtered_checks.remove(check)
                continue

        frequency_delta = parse_frequency(check_config.get('frequency'))
        if not frequency_delta:
            continue

        check_time = probe_for_check_time(config, borg_repository_id, check, archives_check_id)
        if not check_time:
            continue

        # If we've not yet reached the time when the frequency dictates we're ready for another
        # check, skip this check.
        if datetime_now() < check_time + frequency_delta:
            remaining = check_time + frequency_delta - datetime_now()
            logger.info(
                f'Skipping {check} check due to configured frequency; {remaining} until next check (use --force to check anyway)',
            )
            filtered_checks.remove(check)

    return tuple(filtered_checks)


def make_archives_check_id(archive_filter_flags):
    '''
    Given a sequence of flags to filter archives, return a unique hash corresponding to those
    particular flags. If there are no flags, return None.
    '''
    if not archive_filter_flags:
        return None

    return hashlib.sha256(' '.join(archive_filter_flags).encode()).hexdigest()


def make_check_time_path(config, borg_repository_id, check_type, archives_check_id=None):
    '''
    Given a configuration dict, a Borg repository ID, the name of a check type ("repository",
    "archives", etc.), and a unique hash of the archives filter flags, return a path for recording
    that check's time (the time of that check last occurring).
    '''
    borgmatic_state_directory = borgmatic.config.paths.get_borgmatic_state_directory(config)

    if check_type in {'archives', 'data'}:
        return os.path.join(
            borgmatic_state_directory,
            'checks',
            borg_repository_id,
            check_type,
            archives_check_id if archives_check_id else 'all',
        )

    return os.path.join(
        borgmatic_state_directory,
        'checks',
        borg_repository_id,
        check_type,
    )


def write_check_time(path):  # pragma: no cover
    '''
    Record a check time of now as the modification time of the given path.
    '''
    logger.debug(f'Writing check time at {path}')

    os.makedirs(os.path.dirname(path), mode=0o700, exist_ok=True)
    pathlib.Path(path).touch(mode=0o600)


def read_check_time(path):
    '''
    Return the check time based on the modification time of the given path. Return None if the path
    doesn't exist.
    '''
    logger.debug(f'Reading check time from {path}')

    try:
        return datetime.datetime.fromtimestamp(os.stat(path).st_mtime)  # noqa: DTZ006
    except FileNotFoundError:
        return None


def probe_for_check_time(config, borg_repository_id, check, archives_check_id):
    '''
    Given a configuration dict, a Borg repository ID, the name of a check type ("repository",
    "archives", etc.), and a unique hash of the archives filter flags, return the corresponding
    check time or None if such a check time does not exist.

    When the check type is "archives" or "data", this function probes two different paths to find
    the check time, e.g.:

      ~/.borgmatic/checks/1234567890/archives/9876543210
      ~/.borgmatic/checks/1234567890/archives/all

    ... and returns the maximum modification time of the files found (if any). The first path
    represents a more specific archives check time (a check on a subset of archives), and the second
    is a fallback to the last "all" archives check.

    For other check types, this function reads from a single check time path, e.g.:

      ~/.borgmatic/checks/1234567890/repository
    '''
    check_times = (
        read_check_time(group[0])
        for group in itertools.groupby(
            (
                make_check_time_path(config, borg_repository_id, check, archives_check_id),
                make_check_time_path(config, borg_repository_id, check),
            ),
        )
    )

    try:
        return max(check_time for check_time in check_times if check_time)
    except ValueError:
        return None


def upgrade_check_times(config, borg_repository_id):
    '''
    Given a configuration dict and a Borg repository ID, upgrade any corresponding check times on
    disk from old-style paths to new-style paths.

    One upgrade performed is moving the checks directory from:

      {borgmatic_source_directory}/checks (e.g., ~/.borgmatic/checks)

    to:

      {borgmatic_state_directory}/checks (e.g. ~/.local/state/borgmatic)

    Another upgrade is renaming an archive or data check path that looks like:

      {borgmatic_state_directory}/checks/1234567890/archives

    to:

      {borgmatic_state_directory}/checks/1234567890/archives/all
    '''
    borgmatic_source_checks_path = os.path.join(
        borgmatic.config.paths.get_borgmatic_source_directory(config),
        'checks',
    )
    borgmatic_state_path = borgmatic.config.paths.get_borgmatic_state_directory(config)
    borgmatic_state_checks_path = os.path.join(borgmatic_state_path, 'checks')

    if os.path.exists(borgmatic_source_checks_path) and not os.path.exists(
        borgmatic_state_checks_path,
    ):
        logger.debug(
            f'Upgrading archives check times directory from {borgmatic_source_checks_path} to {borgmatic_state_checks_path}',
        )
        os.makedirs(borgmatic_state_path, mode=0o700, exist_ok=True)
        shutil.move(borgmatic_source_checks_path, borgmatic_state_checks_path)

    for check_type in ('archives', 'data'):
        new_path = make_check_time_path(config, borg_repository_id, check_type, 'all')
        old_path = os.path.dirname(new_path)
        temporary_path = f'{old_path}.temp'

        if not os.path.isfile(old_path) and not os.path.isfile(temporary_path):
            continue

        logger.debug(f'Upgrading archives check time file from {old_path} to {new_path}')

        with contextlib.suppress(FileNotFoundError):
            shutil.move(old_path, temporary_path)

        os.mkdir(old_path)
        shutil.move(temporary_path, new_path)


def collect_spot_check_source_paths(
    repository,
    config,
    local_borg_version,
    global_arguments,
    local_path,
    remote_path,
    borgmatic_runtime_directory,
    bootstrap_config_paths,
):
    '''
    Given a repository configuration dict, a configuration dict, the local Borg version, global
    arguments as an argparse.Namespace instance, the local Borg path, the remote Borg path, and the
    bootstrap configuration paths as read from an archive's manifest, collect the source paths that
    Borg would use in an actual create (but only include files). As part of this, include the
    bootstrap configuration paths, so that any configuration files included in the archive to
    support bootstrapping are also spot checked.
    '''
    stream_processes = any(
        borgmatic.hooks.dispatch.call_hooks(
            'use_streaming',
            config,
            borgmatic.hooks.dispatch.Hook_type.DATA_SOURCE,
        ).values(),
    )
    working_directory = borgmatic.config.paths.get_working_directory(config)

    (create_flags, create_positional_arguments, _) = borgmatic.borg.create.make_base_create_command(
        dry_run=True,
        repository_path=repository['path'],
        # Omit "progress" because it interferes with "list_details".
        config=dict(config, progress=False, list_details=True),
        patterns=borgmatic.actions.pattern.process_patterns(
            borgmatic.actions.pattern.collect_patterns(config)
            + tuple(
                borgmatic.borg.pattern.Pattern(
                    config_path,
                    source=borgmatic.borg.pattern.Pattern_source.INTERNAL,
                )
                for config_path in bootstrap_config_paths
            ),
            config,
            working_directory,
        ),
        local_borg_version=local_borg_version,
        global_arguments=global_arguments,
        borgmatic_runtime_directory=borgmatic_runtime_directory,
        local_path=local_path,
        remote_path=remote_path,
        stream_processes=stream_processes,
    )
    working_directory = borgmatic.config.paths.get_working_directory(config)

    path_lines = borgmatic.execute.execute_command_and_capture_output(
        create_flags + create_positional_arguments,
        capture_stderr=True,
        environment=borgmatic.borg.environment.make_environment(config),
        working_directory=working_directory,
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )

    paths = (
        path_line.split(' ', 1)[1]
        for path_line in path_lines
        if path_line and path_line.startswith(('- ', '+ '))
    )

    return tuple(
        # Use dict.fromkeys() to deduplicate file paths, which are present in Borg's dry run output
        # when there are overlapping source patterns. For instance, if both "/foo" and
        # "/foo/file.txt" are in configured patterns, then "/foo/file.txt" will show up in Borg's
        # dry run output twice.
        dict.fromkeys(
            path for path in paths if os.path.isfile(os.path.join(working_directory or '', path))
        )
    )


BORG_DIRECTORY_FILE_TYPE = 'd'
BORG_PIPE_FILE_TYPE = 'p'


def collect_spot_check_archive_paths(
    repository,
    archive,
    config,
    local_borg_version,
    global_arguments,
    local_path,
    remote_path,
    borgmatic_runtime_directory,
):
    '''
    Given a repository configuration dict, the name of the latest archive, a configuration dict, the
    local Borg version, global arguments as an argparse.Namespace instance, the local Borg path, the
    remote Borg path, and the borgmatic runtime directory, collect the paths from the given archive
    (but only include files and symlinks and exclude borgmatic runtime directories).

    These paths do not have a leading slash, as that's how Borg stores them. As a result, we don't
    know whether they came from absolute or relative source directories.
    '''
    borgmatic_source_directory = borgmatic.config.paths.get_borgmatic_source_directory(config)

    return tuple(
        entry['path']
        for entry in borgmatic.borg.list.capture_archive_listing(
            repository['path'],
            archive,
            config,
            local_borg_version,
            global_arguments,
            local_path=local_path,
            remote_path=remote_path,
        )
        if entry['type'] not in {BORG_DIRECTORY_FILE_TYPE, BORG_PIPE_FILE_TYPE}
        if pathlib.Path('borgmatic') not in pathlib.Path(entry['path']).parents
        if pathlib.Path(borgmatic_source_directory.lstrip(os.path.sep))
        not in pathlib.Path(entry['path']).parents
        if pathlib.Path(borgmatic_runtime_directory.lstrip(os.path.sep))
        not in pathlib.Path(entry['path']).parents
    )


SAMPLE_PATHS_SUBSET_COUNT = 5000


def compare_spot_check_hashes(
    repository,
    archive,
    config,
    local_borg_version,
    global_arguments,
    local_path,
    remote_path,
    source_paths,
):
    '''
    Given a repository configuration dict, the name of the latest archive, a configuration dict, the
    local Borg version, global arguments as an argparse.Namespace instance, the local Borg path, the
    remote Borg path, and spot check source paths, compare the hashes for a sampling of the source
    paths with hashes from corresponding paths in the given archive. Return a sequence of the paths
    that fail that hash comparison.
    '''
    # Based on the configured sample percentage, come up with a list of random sample files from the
    # source directories.
    spot_check_config = next(check for check in config['checks'] if check['name'] == 'spot')
    sample_count = max(
        int(len(source_paths) * (min(spot_check_config['data_sample_percentage'], 100) / 100)),
        1,
    )
    source_sample_paths = tuple(random.SystemRandom().sample(source_paths, sample_count))
    working_directory = borgmatic.config.paths.get_working_directory(config)
    hashable_source_sample_path = {
        source_path
        for source_path in source_sample_paths
        for full_source_path in (os.path.join(working_directory or '', source_path),)
        if os.path.exists(full_source_path)
        if not os.path.islink(full_source_path)
    }
    logger.debug(
        f'Sampling {sample_count} source paths (~{spot_check_config["data_sample_percentage"]}%) for spot check',
    )

    source_sample_paths_iterator = iter(source_sample_paths)
    source_hashes = {}
    archive_hashes = {}

    # Only hash a few thousand files at a time (a subset of the total paths) to avoid an "Argument
    # list too long" shell error.
    while True:
        # Hash each file in the sample paths (if it exists).
        source_sample_paths_subset = tuple(
            itertools.islice(source_sample_paths_iterator, SAMPLE_PATHS_SUBSET_COUNT),
        )

        if not source_sample_paths_subset:
            break

        hash_paths = tuple(
            path for path in source_sample_paths_subset if path in hashable_source_sample_path
        )

        try:
            hash_lines = borgmatic.execute.execute_command_and_capture_output(
                tuple(
                    shlex.quote(part)
                    for part in shlex.split(spot_check_config.get('xxh64sum_command', 'xxh64sum'))
                )
                + hash_paths,
                working_directory=working_directory,
            )
            source_hashes.update(
                **dict(
                    zip(
                        # xxh64sum rewrites/escapes the paths that it returns alongside its hashes, for
                        # instance if they contain special characters. When that happens, they don't
                        # match the original source paths and therefore hash lookups fail. So when
                        # building this lookup dict, use the original unaltered paths we provided as
                        # input to xxh64sum.
                        hash_paths,
                        (
                            # For some reason, xxh64sum prefixes the hash with a backslash if the path
                            # contains a newline. Work around that.
                            line.split('  ', 1)[0].lstrip('\\')
                            for line in hash_lines
                        ),
                    ),
                    # Represent non-existent files as having empty hashes so the comparison below still
                    # works. Same thing for filesystem links, since Borg produces empty archive hashes
                    # for them.
                    **{
                        path: ''
                        for path in source_sample_paths_subset
                        if path not in hashable_source_sample_path
                    },
                ),
            )
        except subprocess.CalledProcessError:
            # This can happen if a file we planned to hash gets deleted right before we try to hash
            # it. Falling back to individual file hashing allows us to find and mark just the
            # file(s) with problems instead of failing the whole batch.
            logger.warning(
                'Bulk source path hashing failed for this batch; falling back to individual file hashing'
            )

            for hash_path in hash_paths:
                try:
                    hash_lines = borgmatic.execute.execute_command_and_capture_output(
                        (
                            *(
                                shlex.quote(part)
                                for part in shlex.split(
                                    spot_check_config.get('xxh64sum_command', 'xxh64sum')
                                )
                            ),
                            hash_path,
                        ),
                        working_directory=working_directory,
                    )
                    source_hashes[hash_path] = next(hash_lines).split('  ', 1)[0].lstrip('\\')
                except (subprocess.CalledProcessError, StopIteration):  # noqa: PERF203
                    logger.warning(
                        f'Source path hashing failed for {hash_path}; treating as missing'
                    )
                    source_hashes[hash_path] = ''

        # Get the hash for each file in the archive.
        for entry in borgmatic.borg.list.capture_archive_listing(
            repository['path'],
            archive,
            config,
            local_borg_version,
            global_arguments,
            list_paths=source_sample_paths_subset,
            path_format='{xxh64}{path}{linktarget}{target}',
            local_path=local_path,
            remote_path=remote_path,
        ):
            if not entry:
                continue

            # Borg can't get hashes of stored hard links. So if this is a hard link path (and not
            # deemed as the "original" by Borg), then skip hashing of it.
            if entry.get('linktarget') or entry.get('target'):
                source_hashes.pop(os.path.join('/', entry['path']), None)
                continue

            archive_hashes[entry['path']] = entry['xxh64']

    # Compare the source hashes with the archive hashes to see how many match.
    failing_paths = []

    for path, source_hash in source_hashes.items():
        archive_hash = archive_hashes.get(path.lstrip(os.path.sep))

        if archive_hash is not None and archive_hash == source_hash:
            continue

        failing_paths.append(path)

    return tuple(failing_paths)


MAX_SPOT_CHECK_PATHS_LENGTH = 1000


def spot_check(
    repository,
    config,
    local_borg_version,
    global_arguments,
    local_path,
    remote_path,
    borgmatic_runtime_directory,
):
    '''
    Given a repository dict, a loaded configuration dict, the local Borg version, global arguments
    as an argparse.Namespace instance, the local Borg path, the remote Borg path, and the borgmatic
    runtime directory, perform a spot check for the latest archive in the given repository.

    A spot check compares file counts and also the hashes for a random sampling of source files on
    disk to those stored in the latest archive. If any differences are beyond configured tolerances,
    then the check fails.
    '''
    try:
        spot_check_config = next(
            check for check in config.get('checks', ()) if check.get('name') == 'spot'
        )
    except StopIteration:
        raise ValueError('Cannot run spot check because it is unconfigured')

    if spot_check_config['data_tolerance_percentage'] > spot_check_config['data_sample_percentage']:
        raise ValueError(
            'The data_tolerance_percentage must be less than or equal to the data_sample_percentage',
        )

    archive = borgmatic.borg.repo_list.resolve_archive_name(
        repository['path'],
        'latest',
        config,
        local_borg_version,
        global_arguments,
        local_path,
        remote_path,
    )
    logger.debug(f'Using archive {archive} for spot check')

    source_paths = collect_spot_check_source_paths(
        repository,
        config,
        local_borg_version,
        global_arguments,
        local_path,
        remote_path,
        borgmatic_runtime_directory,
        bootstrap_config_paths=borgmatic.actions.config.bootstrap.load_config_paths_from_archive(
            repository['path'],
            archive,
            config,
            local_borg_version,
            global_arguments,
            borgmatic_runtime_directory,
        ),
    )
    logger.debug(f'{len(source_paths)} total source paths for spot check')

    archive_paths = collect_spot_check_archive_paths(
        repository,
        archive,
        config,
        local_borg_version,
        global_arguments,
        local_path,
        remote_path,
        borgmatic_runtime_directory,
    )
    logger.debug(f'{len(archive_paths)} total archive paths for spot check')

    if len(source_paths) == 0:
        truncated_archive_paths = textwrap.shorten(
            ', '.join(set(archive_paths)) or 'none',
            width=MAX_SPOT_CHECK_PATHS_LENGTH,
            placeholder=' ...',
        )
        logger.debug(f'Paths in latest archive but not source paths: {truncated_archive_paths}')
        raise ValueError(
            'Spot check failed; there are no source paths to compare against the archive',
        )

    # Calculate the percentage delta between the source paths count and the archive paths count, and
    # compare that delta to the configured count tolerance percentage.
    count_delta_percentage = abs(len(source_paths) - len(archive_paths)) / len(source_paths) * 100

    if count_delta_percentage > spot_check_config['count_tolerance_percentage']:
        rootless_source_paths = {path.lstrip(os.path.sep) for path in source_paths}
        truncated_exclusive_source_paths = textwrap.shorten(
            ', '.join(rootless_source_paths - set(archive_paths)) or 'none',
            width=MAX_SPOT_CHECK_PATHS_LENGTH,
            placeholder=' ...',
        )
        truncated_exclusive_archive_paths = textwrap.shorten(
            ', '.join(set(archive_paths) - rootless_source_paths) or 'none',
            width=MAX_SPOT_CHECK_PATHS_LENGTH,
            placeholder=' ...',
        )
        raise ValueError(
            f'Spot check failed\n{count_delta_percentage:.2f}% file count delta between source paths ({len(source_paths)} total) and latest archive ({len(archive_paths)} total); tolerance is {spot_check_config["count_tolerance_percentage"]}%\nOnly in source paths: {truncated_exclusive_source_paths}\nOnly in latest archive: {truncated_exclusive_archive_paths}',
        )

    failing_paths = compare_spot_check_hashes(
        repository,
        archive,
        config,
        local_borg_version,
        global_arguments,
        local_path,
        remote_path,
        source_paths,
    )

    # Error if the percentage of failing hashes exceeds the configured tolerance percentage.
    logger.debug(f'{len(failing_paths)} non-matching spot check hashes')
    data_tolerance_percentage = spot_check_config['data_tolerance_percentage']
    failing_percentage = (len(failing_paths) / len(source_paths)) * 100

    if failing_percentage > data_tolerance_percentage:
        truncated_failing_paths = textwrap.shorten(
            ', '.join(failing_paths),
            width=MAX_SPOT_CHECK_PATHS_LENGTH,
            placeholder=' ...',
        )
        raise ValueError(
            f'Spot check failed\n{failing_percentage:.2f}% of source paths ({len(failing_paths)} out of {len(source_paths)} checked) with data not matching the latest archive; tolerance is {data_tolerance_percentage}%\nSource paths with non-matching data: {truncated_failing_paths}',
        )

    logger.info(
        f'Spot check passed with a {count_delta_percentage:.2f}% file count delta and a {failing_percentage:.2f}% file data delta',
    )


def run_check(
    config_filename,
    repository,
    config,
    local_borg_version,
    check_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "check" action for the given repository.

    Raise ValueError if the Borg repository ID cannot be determined.
    '''
    logger.info('Running consistency checks')

    repository_id = borgmatic.borg.check.get_repository_id(
        repository['path'],
        config,
        local_borg_version,
        global_arguments,
        local_path=local_path,
        remote_path=remote_path,
    )
    upgrade_check_times(config, repository_id)
    configured_checks = parse_checks(config, check_arguments.only_checks)
    archive_filter_flags = borgmatic.borg.check.make_archive_filter_flags(
        local_borg_version,
        config,
        configured_checks,
        check_arguments,
    )
    archives_check_id = make_archives_check_id(archive_filter_flags)
    checks = filter_checks_on_frequency(
        config,
        repository_id,
        configured_checks,
        check_arguments.force,
        archives_check_id,
    )
    borg_specific_checks = set(checks).intersection({'repository', 'archives', 'data'})

    if borg_specific_checks:
        borgmatic.borg.check.check_archives(
            repository['path'],
            config,
            local_borg_version,
            check_arguments,
            global_arguments,
            borg_specific_checks,
            archive_filter_flags,
            local_path=local_path,
            remote_path=remote_path,
        )
        for check in borg_specific_checks:
            write_check_time(make_check_time_path(config, repository_id, check, archives_check_id))

    if 'extract' in checks:
        logger.info('Running extract check')
        borgmatic.borg.extract.extract_last_archive_dry_run(
            config,
            local_borg_version,
            global_arguments,
            repository['path'],
            config.get('lock_wait'),
            local_path,
            remote_path,
        )
        write_check_time(make_check_time_path(config, repository_id, 'extract'))

    if 'spot' in checks:
        logger.info('Running spot check')
        with borgmatic.config.paths.Runtime_directory(config) as borgmatic_runtime_directory:
            spot_check(
                repository,
                config,
                local_borg_version,
                global_arguments,
                local_path,
                remote_path,
                borgmatic_runtime_directory,
            )

        write_check_time(make_check_time_path(config, repository_id, 'spot'))



import argparse
import logging
import shlex

import borgmatic.config.paths
import borgmatic.logger
from borgmatic.borg import environment, feature, flags
from borgmatic.execute import execute_command_and_capture_output

logger = logging.getLogger(__name__)


def make_info_command(
    repository_path,
    config,
    local_borg_version,
    info_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Given a local or remote repository path, a configuration dict, the local Borg version, the
    arguments to the info action as an argparse.Namespace, and global arguments, return a command
    as a tuple to display summary information for archives in the repository.
    '''
    extra_borg_options = config.get('extra_borg_options', {}).get('info', '')

    return (
        (local_path, 'info')
        + (
            ('--info',)
            if logger.getEffectiveLevel() == logging.INFO and not info_arguments.json
            else ()
        )
        + (
            ('--debug', '--show-rc')
            if logger.isEnabledFor(logging.DEBUG) and not info_arguments.json
            else ()
        )
        + flags.make_flags('remote-path', remote_path)
        + flags.make_flags('umask', config.get('umask'))
        + ('--log-json',)
        + flags.make_flags('lock-wait', config.get('lock_wait'))
        + (
            (
                flags.make_flags('match-archives', f'sh:{info_arguments.prefix}*')
                if feature.available(feature.Feature.MATCH_ARCHIVES, local_borg_version)
                else flags.make_flags('glob-archives', f'{info_arguments.prefix}*')
            )
            if info_arguments.prefix
            else (
                flags.make_match_archives_flags(
                    info_arguments.archive or config.get('match_archives'),
                    config.get('archive_name_format'),
                    local_borg_version,
                )
            )
        )
        + flags.make_flags_from_arguments(
            info_arguments,
            excludes=('repository', 'archive', 'prefix', 'match_archives'),
        )
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + flags.make_repository_flags(repository_path, local_borg_version)
    )


def display_archives_info(
    repository_path,
    config,
    local_borg_version,
    info_arguments,
    global_arguments,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a local or remote repository path, a configuration dict, the local Borg version, the
    arguments to the info action as an argparse.Namespace, and global arguments, display summary
    information for Borg archives in the repository or return JSON summary information.
    '''
    borgmatic.logger.add_custom_log_levels()

    main_command = make_info_command(
        repository_path,
        config,
        local_borg_version,
        info_arguments,
        global_arguments,
        local_path,
        remote_path,
    )
    json_command = make_info_command(
        repository_path,
        config,
        local_borg_version,
        argparse.Namespace(**dict(info_arguments.__dict__, json=True)),
        global_arguments,
        local_path,
        remote_path,
    )
    borg_exit_codes = config.get('borg_exit_codes')
    working_directory = borgmatic.config.paths.get_working_directory(config)

    if info_arguments.json:
        return '\n'.join(
            execute_command_and_capture_output(
                json_command,
                environment=environment.make_environment(config),
                working_directory=working_directory,
                borg_local_path=local_path,
                borg_exit_codes=borg_exit_codes,
            )
        )

    output_lines = tuple(
        execute_command_and_capture_output(
            main_command,
            output_log_level=logging.ANSWER,
            environment=environment.make_environment(config),
            working_directory=working_directory,
            borg_local_path=local_path,
            borg_exit_codes=borg_exit_codes,
        )
    )

    flags.warn_for_aggressive_archive_flags(main_command, output_lines)

    return None



import json
import logging
import sys

import borgmatic.config.generate
import borgmatic.logger

logger = logging.getLogger(__name__)


def run_show(show_arguments, configs):
    '''
    Given the show arguments as an argparse.Namespace instance and a dict of configuration filename
    to corresponding parsed configuration, run the "show" action. That consists of rendering and
    logging the computed configuration as YAML, separating the configuration for each file with
    "---".

    If show_arguments.option is set, limit the results to the value of that single option. If
    show_arguments.json is True, render the results as JSON with one array element per configuration
    file.
    '''
    borgmatic.logger.add_custom_log_levels()

    if show_arguments.json:
        sys.stdout.write(
            json.dumps(
                [
                    config.get(show_arguments.option) if show_arguments.option else config
                    for config in configs.values()
                ]
            )
        )

        return

    for config in configs.values():
        if len(configs) > 1:
            logger.answer('---')

        logger.answer(
            borgmatic.config.generate.render_configuration(
                config.get(show_arguments.option) if show_arguments.option else config
            ).rstrip()
        )



import logging

import requests

import borgmatic.hooks.credential.parse

logger = logging.getLogger(__name__)


TIMEOUT_SECONDS = 10


def initialize_monitor(
    ping_url,
    config,
    config_filename,
    monitoring_log_level,
    dry_run,
):  # pragma: no cover
    '''
    No initialization is necessary for this monitor.
    '''


def send_zabbix_request(server, headers, data):
    '''
    Given a Zabbix server URL, HTTP headers as a dict, and valid Zabbix JSON payload data as a dict,
    send a request to the Zabbix server via API.

    Return the response "result" value or None.
    '''
    logging.getLogger('urllib3').setLevel(logging.ERROR)

    logger.debug(f'Sending a "{data["method"]}" request to the Zabbix server')

    try:
        response = requests.post(server, headers=headers, json=data, timeout=TIMEOUT_SECONDS)

        if not response.ok:
            response.raise_for_status()
    except requests.exceptions.RequestException as error:
        logger.warning(f'Zabbix error: {error}')

        return None

    try:
        result = response.json().get('result')
        error_message = result['data'][0]['error']
    except requests.exceptions.JSONDecodeError:
        logger.warning('Zabbix error: Cannot parse API response')

        return None
    except (TypeError, KeyError, IndexError):
        return result
    else:
        logger.warning(f'Zabbix error: {error_message}')

        return None


def ping_monitor(hook_config, config, config_filename, state, monitoring_log_level, dry_run):  # noqa: PLR0911, PLR0912, PLR0915
    '''
    Update the configured Zabbix item using either the itemid, or a host and key.
    If this is a dry run, then don't actually update anything.
    '''

    run_states = hook_config.get('states', ['fail'])

    if state.name.lower() not in run_states:
        return

    dry_run_label = ' (dry run; not actually updating)' if dry_run else ''

    state_config = hook_config.get(
        state.name.lower(),
        {
            'value': state.name.lower(),
        },
    )

    try:
        username = borgmatic.hooks.credential.parse.resolve_credential(
            hook_config.get('username'),
            config,
        )
        password = borgmatic.hooks.credential.parse.resolve_credential(
            hook_config.get('password'),
            config,
        )
        api_key = borgmatic.hooks.credential.parse.resolve_credential(
            hook_config.get('api_key'),
            config,
        )
    except ValueError as error:
        logger.warning(f'Zabbix credential error: {error}')

        return

    server = hook_config.get('server')
    itemid = hook_config.get('itemid')
    host = hook_config.get('host')
    key = hook_config.get('key')
    value = state_config.get('value')
    headers = {'Content-Type': 'application/json-rpc', 'User-Agent': 'borgmatic'}

    logger.info(f'Pinging Zabbix{dry_run_label}')
    logger.debug(f'Using Zabbix URL: {server}')

    # Determine the Zabbix method used to store the value: itemid or host/key
    if itemid is not None:
        logger.info(f'Updating {itemid} on Zabbix')
        data = {
            'jsonrpc': '2.0',
            'method': 'history.push',
            'params': {'itemid': itemid, 'value': value},
            'id': 1,
        }

    elif host is not None and key is not None:
        logger.info(f'Updating Host: "{host}" and Key: "{key}" on Zabbix')
        data = {
            'jsonrpc': '2.0',
            'method': 'history.push',
            'params': {'host': host, 'key': key, 'value': value},
            'id': 1,
        }

    elif host is not None:
        logger.warning('Key missing for Zabbix')

        return
    elif key is not None:
        logger.warning('Host missing for Zabbix')

        return
    else:
        logger.warning('No Zabbix itemid or host/key provided')

        return

    # Determine the authentication method: API key or username/password
    if api_key is not None:
        logger.info('Using API key auth for Zabbix')
        headers['Authorization'] = f'Bearer {api_key}'
    elif username is not None and password is not None:
        logger.info(f'Using user/pass auth with user {username} for Zabbix')
        login_data = {
            'jsonrpc': '2.0',
            'method': 'user.login',
            'params': {'username': username, 'password': password},
            'id': 1,
        }

        if not dry_run:
            result = send_zabbix_request(server, headers, login_data)

            if not result:
                return

            headers['Authorization'] = f'Bearer {result}'
    elif username is not None:
        logger.warning('Password missing for Zabbix authentication')

        return
    elif password is not None:
        logger.warning('Username missing for Zabbix authentication')

        return
    else:
        logger.warning('Authentication data missing for Zabbix')

        return

    if not dry_run:
        send_zabbix_request(server, headers, data)

    if username is not None and password is not None:
        logout_data = {
            'jsonrpc': '2.0',
            'method': 'user.logout',
            'params': [],
            'id': 1,
        }

        if not dry_run:
            send_zabbix_request(server, headers, logout_data)


def destroy_monitor(ping_url_or_uuid, config, monitoring_log_level, dry_run):  # pragma: no cover
    '''
    No destruction is necessary for this monitor.
    '''
